"""
Copyright (c) 2024, Daily
SPDX-License-Identifier: BSD 2-Clause License

backend for https://github.com/rtvi-ai to adapter frontend client-side RTVI  stack
- implements RTVI events and data structures.
- processor orchestration
- config and event update

In addition, there will be future network protocols and standards that make sense to use for real-time AI. (For example, [Media over Quic](https://quic.video/)).
"""

import logging
import asyncio
from typing import Any, Awaitable, Callable, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, PrivateAttr, ValidationError
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor
from apipeline.frames.sys_frames import SystemFrame, CancelFrame, ErrorFrame
from apipeline.frames.control_frames import StartFrame, EndFrame

from achatbot.types.frames.sys_frames import BotInterruptionFrame
from achatbot.types.frames.control_frames import (
    BotStartedSpeakingFrame,
    BotStoppedSpeakingFrame,
    UserStartedSpeakingFrame,
    UserStoppedSpeakingFrame,
)
from achatbot.types.frames.data_frames import (
    Frame,
    FunctionCallResultFrame,
    TranscriptionFrame,
    InterimTranscriptionFrame,
    TransportMessageFrame,
)
from achatbot.processors.aggregators.openai_llm_context import OpenAILLMContext

from dotenv import load_dotenv

load_dotenv(override=True)


RTVI_PROTOCOL_VERSION = "0.1"

ActionResult = Union[bool, int, float, str, list, dict]


class RTVIServiceOption(BaseModel):
    name: str
    type: Literal["bool", "number", "string", "array", "object", "dict", "list"]
    handler: Callable[["RTVIProcessor", str, "RTVIServiceOptionConfig"], Awaitable[None]] = Field(
        exclude=True
    )


class RTVIService(BaseModel):
    name: str
    options: List[RTVIServiceOption]
    _options_dict: Dict[str, RTVIServiceOption] = PrivateAttr(default={})

    def model_post_init(self, __context: Any) -> None:
        self._options_dict = {}
        for option in self.options:
            self._options_dict[option.name] = option
        return super().model_post_init(__context)


class RTVIActionArgumentData(BaseModel):
    name: str
    value: Any


class RTVIActionArgument(BaseModel):
    name: str
    type: Literal["bool", "number", "string", "array", "object", "dict", "list"]


class RTVIAction(BaseModel):
    service: str
    action: str
    handler: Callable[["RTVIProcessor", str, Dict[str, Any]], Awaitable[ActionResult]] = Field(
        exclude=True
    )


#
# Client -> achatbot messages.
#


class RTVIServiceOptionConfig(BaseModel):
    name: str
    value: Any


class RTVIServiceConfig(BaseModel):
    service: str
    options: List[RTVIServiceOptionConfig]


class RTVIConfig(BaseModel):
    config_list: List[RTVIServiceConfig]
    _arguments_dict: Dict[str, Any] = PrivateAttr(default={})

    def model_post_init(self, __context: Any) -> None:
        self._arguments_dict = {}
        for conf in self.config_list:
            opt_dict = {}
            for opt in conf.options:
                opt_dict[opt.name] = opt.value
            self._arguments_dict[conf.service] = opt_dict
        return super().model_post_init(__context)


class RTVIActionRunArgument(BaseModel):
    name: str
    value: Any


class RTVIActionRun(BaseModel):
    service: str
    action: str
    arguments: Optional[List[RTVIActionRunArgument]] = None


class RTVIMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: str
    id: str
    data: Optional[Dict[str, Any]] = None


#
# achatbot -> Client responses and messages.
#


class RTVIErrorResponseData(BaseModel):
    error: str


class RTVIErrorResponse(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["error-response"] = "error-response"
    id: str
    data: RTVIErrorResponseData


class RTVIErrorData(BaseModel):
    error: str
    fatal: bool


class RTVIError(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["error"] = "error"
    data: RTVIErrorData


class RTVIDescribeConfigData(BaseModel):
    config: List[RTVIService]


class RTVIDescribeConfig(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["config-available"] = "config-available"
    id: str
    data: RTVIDescribeConfigData


class RTVIDescribeActionsData(BaseModel):
    actions: List[RTVIAction]


class RTVIDescribeActions(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["actions-available"] = "actions-available"
    id: str
    data: RTVIDescribeActionsData


class RTVIConfigResponse(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["config"] = "config"
    id: str
    data: RTVIConfig


class RTVIActionResponseData(BaseModel):
    result: ActionResult


class RTVIActionResponse(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["action-response"] = "action-response"
    id: str
    data: RTVIActionResponseData


class RTVIBotReadyData(BaseModel):
    version: str
    config: List[RTVIServiceConfig]


class RTVIBotReady(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["bot-ready"] = "bot-ready"
    id: str
    data: RTVIBotReadyData


class RTVILLMFunctionCallMessageData(BaseModel):
    function_name: str
    tool_call_id: str
    args: dict


class RTVILLMFunctionCallMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["llm-function-call"] = "llm-function-call"
    data: RTVILLMFunctionCallMessageData


class RTVILLMFunctionCallStartMessageData(BaseModel):
    function_name: str


class RTVILLMFunctionCallStartMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["llm-function-call-start"] = "llm-function-call-start"
    data: RTVILLMFunctionCallStartMessageData


class RTVILLMFunctionCallResultData(BaseModel):
    function_name: str
    tool_call_id: str
    arguments: dict
    result: dict


class RTVITranscriptionMessageData(BaseModel):
    text: str
    user_id: str
    timestamp: str
    final: bool


class RTVITranscriptionMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["user-transcription"] = "user-transcription"
    data: RTVITranscriptionMessageData


class RTVIUserStartedSpeakingMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["user-started-speaking"] = "user-started-speaking"


class RTVIUserStoppedSpeakingMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["user-stopped-speaking"] = "user-stopped-speaking"


class RTVIBotStartedSpeakingMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["bot-started-speaking"] = "bot-started-speaking"


class RTVIBotStoppedSpeakingMessage(BaseModel):
    label: Literal["rtvi-ai"] = "rtvi-ai"
    type: Literal["bot-stopped-speaking"] = "bot-stopped-speaking"


class RTVIProcessorParams(BaseModel):
    send_bot_ready: bool = True


class RTVIProcessor(FrameProcessor):
    def __init__(
        self,
        *,
        config: RTVIConfig = RTVIConfig(config_list=[]),
        params: RTVIProcessorParams = RTVIProcessorParams(),
    ):
        super().__init__()
        self._config = config
        self._params = params

        self._pipeline: FrameProcessor | None = None
        self._pipeline_started = False

        self._client_ready = False
        self._client_ready_id = ""

        self._registered_actions: Dict[str, RTVIAction] = {}
        self._registered_services: Dict[str, RTVIService] = {}

        self._push_frame_task = self.get_event_loop().create_task(self._push_frame_task_handler())
        self._push_queue = asyncio.Queue()

        self._message_task = self.get_event_loop().create_task(self._message_task_handler())
        self._message_queue = asyncio.Queue()

        self._curr_rtvi_message: RTVIMessage = None

    @property
    def curr_rtvi_meesage(self):
        return self._curr_rtvi_message

    def register_action(self, action: RTVIAction):
        id = self._action_id(action.service, action.action)
        self._registered_actions[id] = action

    def register_actions(self, actions: List[RTVIAction]):
        for action in actions:
            self.register_action(action)

    def register_service(self, service: RTVIService):
        self._registered_services[service.name] = service

    def register_services(self, services: List[RTVIService]):
        for service in services:
            self.register_service(service)

    async def interrupt_bot(self):
        await self.push_frame(BotInterruptionFrame(), FrameDirection.UPSTREAM)

    async def send_error(self, error: str):
        message = RTVIError(data=RTVIErrorData(error=error, fatal=False))
        await self._push_transport_message(message)

    async def set_client_ready(self):
        if not self._client_ready:
            self._client_ready = True
            await self._maybe_send_bot_ready()

    async def handle_function_call(
        self,
        function_name: str,
        tool_call_id: str,
        arguments: dict,
        llm: FrameProcessor,
        context: OpenAILLMContext,
        result_callback,
    ):
        fn = RTVILLMFunctionCallMessageData(
            function_name=function_name, tool_call_id=tool_call_id, args=arguments
        )
        message = RTVILLMFunctionCallMessage(data=fn)
        await self._push_transport_message(message, exclude_none=False)

    async def handle_function_call_start(
        self, function_name: str, llm: FrameProcessor, context: OpenAILLMContext
    ):
        fn = RTVILLMFunctionCallStartMessageData(function_name=function_name)
        message = RTVILLMFunctionCallStartMessage(data=fn)
        await self._push_transport_message(message, exclude_none=False)

    async def push_frame(self, frame: Frame, direction: FrameDirection = FrameDirection.DOWNSTREAM):
        if isinstance(frame, SystemFrame):
            await super().push_frame(frame, direction)
        else:
            await self._internal_push_frame(frame, direction)

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        # print("rtvi process frame", frame)
        # Specific system frames
        if isinstance(frame, CancelFrame):
            await self._cancel(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, ErrorFrame):
            await self._send_error_frame(frame)
            await self.push_frame(frame, direction)
        # All other system frames
        elif isinstance(frame, SystemFrame):
            await self.push_frame(frame, direction)
        # Control frames
        elif isinstance(frame, StartFrame):
            await self.push_frame(frame, direction)
            await self._start(frame)
        elif isinstance(frame, EndFrame):
            # Push EndFrame before stop(), because stop() waits on the task to
            # finish and the task finishes when EndFrame is processed.
            await self.push_frame(frame, direction)
            await self._stop(frame)
        elif isinstance(frame, UserStartedSpeakingFrame) or isinstance(
            frame, UserStoppedSpeakingFrame
        ):
            await self._handle_interruptions(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, BotStartedSpeakingFrame) or isinstance(
            frame, BotStoppedSpeakingFrame
        ):
            await self._handle_bot_speaking(frame)
            await self.push_frame(frame, direction)
        # Data frames
        elif isinstance(frame, TranscriptionFrame) or isinstance(frame, InterimTranscriptionFrame):
            await self._handle_transcriptions(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, TransportMessageFrame):
            await self._message_queue.put(frame)
        # Other frames
        else:
            await self.push_frame(frame, direction)

    async def cleanup(self):
        if self._pipeline:
            await self._pipeline.cleanup()

    async def _start(self, frame: StartFrame):
        self._pipeline_started = True
        await self._maybe_send_bot_ready()

    async def _stop(self, frame: EndFrame):
        # We need to cancel the message task handler because that one is not
        # processing EndFrames.
        self._message_task.cancel()
        await self._message_task
        await self._push_frame_task

    async def _cancel(self, frame: CancelFrame):
        self._message_task.cancel()
        await self._message_task
        self._push_frame_task.cancel()
        await self._push_frame_task

    async def _internal_push_frame(
        self, frame: Frame | None, direction: FrameDirection | None = FrameDirection.DOWNSTREAM
    ):
        await self._push_queue.put((frame, direction))

    async def _push_frame_task_handler(self):
        running = True
        while running:
            try:
                (frame, direction) = await self._push_queue.get()
                await super().push_frame(frame, direction)
                self._push_queue.task_done()
                running = not isinstance(frame, EndFrame)
            except asyncio.CancelledError:
                break

    async def _push_transport_message(self, model: BaseModel, exclude_none: bool = True):
        frame = TransportMessageFrame(
            message=model.model_dump(exclude_none=exclude_none), urgent=True
        )
        await self.push_frame(frame)

    async def _handle_transcriptions(self, frame: Frame):
        message = None
        if isinstance(frame, TranscriptionFrame):
            message = RTVITranscriptionMessage(
                data=RTVITranscriptionMessageData(
                    text=frame.text, user_id=frame.user_id, timestamp=frame.timestamp, final=True
                )
            )
        elif isinstance(frame, InterimTranscriptionFrame):
            message = RTVITranscriptionMessage(
                data=RTVITranscriptionMessageData(
                    text=frame.text, user_id=frame.user_id, timestamp=frame.timestamp, final=False
                )
            )

        if message:
            await self._push_transport_message(message)

    async def _handle_interruptions(self, frame: Frame):
        message = None
        if isinstance(frame, UserStartedSpeakingFrame):
            message = RTVIUserStartedSpeakingMessage()
        elif isinstance(frame, UserStoppedSpeakingFrame):
            message = RTVIUserStoppedSpeakingMessage()

        if message:
            await self._push_transport_message(message)

    async def _handle_bot_speaking(self, frame: Frame):
        message = None
        if isinstance(frame, BotStartedSpeakingFrame):
            message = RTVIBotStartedSpeakingMessage()
        elif isinstance(frame, BotStoppedSpeakingFrame):
            message = RTVIBotStoppedSpeakingMessage()

        if message:
            await self._push_transport_message(message)

    async def _message_task_handler(self):
        while True:
            try:
                frame = await self._message_queue.get()
                await self._handle_message(frame)
                self._message_queue.task_done()
            except asyncio.CancelledError:
                break

    async def _handle_message(self, frame: TransportMessageFrame):
        logging.debug(f"handle TransportMessageFrame: {frame}")
        try:
            message = RTVIMessage.model_validate(frame.message)
        except ValidationError as e:
            await self.send_error(f"Invalid incoming message: {e}")
            logging.warning(f"Invalid incoming  message: {e}")
            return

        logging.info(f"rtvi server handle message: {message}")
        self._curr_rtvi_message = message
        try:
            match message.type:
                case "client-ready":
                    await self._handle_client_ready(message.id)
                case "describe-actions":
                    await self._handle_describe_actions(message.id)
                case "describe-config":
                    await self._handle_describe_config(message.id)
                case "get-config":
                    await self._handle_get_config(message.id)
                case "update-config":
                    config = RTVIConfig.model_validate(message.data)
                    await self._handle_update_config(message.id, config)
                case "action":
                    action = RTVIActionRun.model_validate(message.data)
                    await self._handle_action(message.id, action)
                case "llm-function-call-result":
                    data = RTVILLMFunctionCallResultData.model_validate(message.data)
                    await self._handle_function_call_result(data)

                case _:
                    await self._send_error_response(message.id, f"Unsupported type {message.type}")

        except ValidationError as e:
            await self._send_error_response(message.id, f"Invalid incoming message: {e}")
            logging.warning(f"Invalid incoming  message: {e}")
        except Exception as e:
            await self._send_error_response(message.id, f"Exception processing message: {e}")
            logging.warning(f"Exception processing message: {e}", exc_info=True)

    async def _handle_client_ready(self, request_id: str):
        self._client_ready = True
        self._client_ready_id = request_id
        await self._maybe_send_bot_ready()

    async def _handle_describe_config(self, request_id: str):
        services = list(self._registered_services.values())
        message = RTVIDescribeConfig(id=request_id, data=RTVIDescribeConfigData(config=services))
        await self._push_transport_message(message)

    async def _handle_describe_actions(self, request_id: str):
        actions = list(self._registered_actions.values())
        message = RTVIDescribeActions(id=request_id, data=RTVIDescribeActionsData(actions=actions))
        await self._push_transport_message(message)

    async def _handle_get_config(self, request_id: str):
        message = RTVIConfigResponse(id=request_id, data=self._config)
        await self._push_transport_message(message)

    def _update_config_option(self, service: str, config: RTVIServiceOptionConfig):
        for service_config in self._config.config_list:
            if service_config.service == service:
                for option_config in service_config.options:
                    if option_config.name == config.name:
                        option_config.value = config.value
                        return
                # If we couldn't find a value for this config, we simply need to
                # add it.
                service_config.options.append(config)

    async def _update_service_config(self, config: RTVIServiceConfig):
        if config.service not in self._registered_services:
            return
        service = self._registered_services[config.service]
        for option in config.options:
            if option.name in service._options_dict:
                handler = service._options_dict[option.name].handler
                await handler(self, service.name, option)
            else:
                logging.info(f"{option.name} not handler")
            self._update_config_option(service.name, option)

    async def _update_config(self, data: RTVIConfig):
        for service_config in data.config_list:
            await self._update_service_config(service_config)

    async def _handle_update_config(self, request_id: str, data: RTVIConfig):
        # NOTE: The bot might be talking while we receive a new
        # config. Let's interrupt it for now and update the config. Another
        # solution is to wait until the bot stops speaking and then apply the
        # config, but this definitely is more complicated to achieve.
        await self.interrupt_bot()
        await self._update_config(data)
        await self._handle_get_config(request_id)

    async def _handle_function_call_result(self, data):
        frame = FunctionCallResultFrame(
            function_name=data.function_name,
            tool_call_id=data.tool_call_id,
            arguments=data.arguments,
            result=data.result,
        )
        await self.push_frame(frame)

    async def _handle_action(self, request_id: str, data: RTVIActionRun):
        action_id = self._action_id(data.service, data.action)
        if action_id not in self._registered_actions:
            await self._send_error_response(request_id, f"Action {action_id} not registered")
            return
        action = self._registered_actions[action_id]
        arguments = {}
        if data.arguments:
            for arg in data.arguments:
                arguments[arg.name] = arg.value
        result = await action.handler(self, action.service, arguments)
        message = RTVIActionResponse(id=request_id, data=RTVIActionResponseData(result=result))
        await self._push_transport_message(message)

    async def _maybe_send_bot_ready(self):
        if self._pipeline_started and self._client_ready:
            await self._send_bot_ready()
            await self._update_config(self._config)

    async def _send_bot_ready(self):
        if not self._params.send_bot_ready:
            return

        message = RTVIBotReady(
            id=self._client_ready_id,
            data=RTVIBotReadyData(version=RTVI_PROTOCOL_VERSION, config=self._config.config_list),
        )
        await self._push_transport_message(message)

    async def _send_error_frame(self, frame: ErrorFrame):
        message = RTVIError(data=RTVIErrorData(error=frame.error, fatal=frame.fatal))
        await self._push_transport_message(message)

    async def _send_error_response(self, id: str, error: str):
        message = RTVIErrorResponse(id=id, data=RTVIErrorResponseData(error=error))
        await self._push_transport_message(message)

    def _action_id(self, service: str, action: str) -> str:
        return f"{service}:{action}"



import modal
import os


class ContainerRuntimeConfig:
    images = {
        # image key name is shot, lenght less 10
        "default": (
            modal.Image.debian_slim(python_version="3.11")
            .apt_install("git", "git-lfs", "ffmpeg")
            .pip_install(
                [
                    "achatbot["
                    "fastapi_bot_server,"
                    "livekit,livekit-api,daily,agora,"
                    "silero_vad_analyzer,daily_langchain_rag_bot,"
                    "sense_voice_asr,deepgram_asr_processor,"
                    "openai_llm_processor,google_llm_processor,litellm_processor,"
                    "tts_edge,"
                    "deep_translator,together_ai,"
                    "queue"
                    "]~=0.0.8.3",
                    "huggingface_hub[hf_transfer]==0.24.7",
                ],
                extra_index_url="https://pypi.org/simple/",
            )
            .env(
                {
                    "HF_HUB_ENABLE_HF_TRANSFER": "1",
                    "ACHATBOT_PKG": "1",
                    "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
                    "IMAGE_NAME": os.getenv("IMAGE_NAME", "default"),
                    # asr module engine TAG, default whisper_timestamped_asr
                    "ASR_TAG": "sense_voice_asr",
                    "ASR_LANG": "zn",
                    "ASR_MODEL_NAME_OR_PATH": "/root/.achatbot/models/FunAudioLLM/SenseVoiceSmall",
                    # llm processor model, default:google gemini_flash_latest
                    "GOOGLE_LLM_MODEL": "gemini-1.5-flash-latest",
                    # tts module engine TAG,default tts_edge
                    "TTS_TAG": "tts_edge",
                }
            )
        ),
        "qwen": (
            modal.Image.debian_slim(python_version="3.11")
            .apt_install("git", "git-lfs", "ffmpeg")
            .pip_install(
                [
                    "achatbot["
                    "fastapi_bot_server,"
                    "livekit,livekit-api,daily,agora,"
                    "silero_vad_analyzer,daily_langchain_rag_bot,"
                    "sense_voice_asr,deepgram_asr_processor,"
                    "llm_transformers_manual_vision_qwen,"
                    "openai_llm_processor,google_llm_processor,litellm_processor,"
                    "tts_edge,"
                    "deep_translator,together_ai,"
                    "queue"
                    "]~=0.0.8.3",
                    "huggingface_hub[hf_transfer]==0.24.7",
                ],
                extra_index_url="https://pypi.org/simple/",
            )
            .env(
                {
                    "HF_HUB_ENABLE_HF_TRANSFER": "1",
                    "ACHATBOT_PKG": "1",
                    "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
                    "IMAGE_NAME": os.getenv("IMAGE_NAME", "default"),
                    # asr module engine TAG, default whisper_timestamped_asr
                    "ASR_TAG": "sense_voice_asr",
                    "ASR_LANG": "zn",
                    "ASR_MODEL_NAME_OR_PATH": "/root/.achatbot/models/FunAudioLLM/SenseVoiceSmall",
                    # llm processor model, default:google gemini_flash_latest
                    "GOOGLE_LLM_MODEL": "gemini-1.5-flash-latest",
                    "LLM_MODEL_NAME_OR_PATH": f'/root/.achatbot/models/{os.getenv("LLM_MODEL_NAME_OR_PATH", "Qwen/Qwen2-VL-2B-Instruct")}',
                    # tts module engine TAG,default tts_edge
                    "TTS_TAG": "tts_edge",
                }
            )
        ),
        "llama": (
            modal.Image.debian_slim(python_version="3.11")
            .apt_install("git", "git-lfs", "ffmpeg")
            .pip_install(
                [
                    "achatbot["
                    "fastapi_bot_server,"
                    "livekit,livekit-api,daily,agora,"
                    "silero_vad_analyzer,daily_langchain_rag_bot,"
                    "sense_voice_asr,deepgram_asr_processor,"
                    "llm_transformers_manual_vision_llama,"
                    "openai_llm_processor,google_llm_processor,litellm_processor,"
                    "tts_edge,"
                    "deep_translator,together_ai,"
                    "queue"
                    "]~=0.0.8.3",
                    "huggingface_hub[hf_transfer]==0.24.7",
                ],
                extra_index_url="https://pypi.org/simple/",
            )
            .env(
                {
                    "HF_HUB_ENABLE_HF_TRANSFER": "1",
                    "ACHATBOT_PKG": "1",
                    "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
                    "IMAGE_NAME": os.getenv("IMAGE_NAME", "default"),
                    # asr module engine TAG, default whisper_timestamped_asr
                    "ASR_TAG": "sense_voice_asr",
                    "ASR_LANG": "zn",
                    "ASR_MODEL_NAME_OR_PATH": "FunAudioLLM/SenseVoiceSmall",
                    # llm processor model, default:google gemini_flash_latest
                    "GOOGLE_LLM_MODEL": "gemini-1.5-flash-latest",
                    "LLM_MODEL_NAME_OR_PATH": f'/root/.achatbot/models/{os.getenv("LLM_MODEL_NAME_OR_PATH", "unsloth/Llama-3.2-11B-Vision-Instruct")}',
                    # tts module engine TAG,default tts_edge
                    "TTS_TAG": "tts_edge",
                }
            )
        ),
        "janus": (
            modal.Image.debian_slim(python_version="3.11")
            .apt_install("git", "git-lfs", "ffmpeg")
            .pip_install(
                [
                    "achatbot["
                    "fastapi_bot_server,"
                    "livekit,livekit-api,daily,agora,"
                    "silero_vad_analyzer,daily_langchain_rag_bot,"
                    "sense_voice_asr,deepgram_asr_processor,"
                    "llm_transformers_manual_vision_img_janus,"
                    "openai_llm_processor,google_llm_processor,litellm_processor,"
                    "tts_edge,"
                    "deep_translator,together_ai,"
                    "queue"
                    "]~=0.0.8.7",
                    "huggingface_hub[hf_transfer]==0.24.7",
                ],
                extra_index_url="https://pypi.org/simple/",
            )
            .env(
                {
                    "HF_HUB_ENABLE_HF_TRANSFER": "1",
                    "ACHATBOT_PKG": "1",
                    "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
                    "IMAGE_NAME": os.getenv("IMAGE_NAME", "default"),
                    # asr module engine TAG, default whisper_timestamped_asr
                    "ASR_TAG": "sense_voice_asr",
                    "ASR_LANG": "zn",
                    "ASR_MODEL_NAME_OR_PATH": "/root/.achatbot/models/FunAudioLLM/SenseVoiceSmall",
                    # llm processor model, default:google gemini_flash_latest
                    "GOOGLE_LLM_MODEL": "gemini-1.5-flash-latest",
                    "LLM_MODEL_NAME_OR_PATH": f'/root/.achatbot/models/{os.getenv("LLM_MODEL_NAME_OR_PATH", "deepseek-ai/Janus-Pro-1B")}',
                    # tts module engine TAG,default tts_edge
                    "TTS_TAG": "tts_edge",
                }
            )
        ),
    }

    @staticmethod
    def get_img(image_name: str = None):
        image_name = image_name or os.getenv("IMAGE_NAME", "default")
        if image_name not in ContainerRuntimeConfig.images:
            raise Exception(f"image name {image_name} not found")
        print(f"use image:{image_name}")
        return ContainerRuntimeConfig.images[image_name]

    @staticmethod
    def get_app_name(image_name: str = None):
        image_name = image_name or os.getenv("IMAGE_NAME", "default")
        app_name = "fastapi_webrtc_vision_bot"
        if image_name != "default":
            app_name = f"fastapi_webrtc_vision_{image_name}_bot"
        print(f"app_name:{app_name}")
        return app_name

    @staticmethod
    def get_gpu():
        # T4, L4, A10G, A100, H100
        gpu = os.getenv("IMAGE_GPU", None)
        print(f"image_gpu:{gpu}")
        return gpu

    @staticmethod
    def get_allow_concurrent_inputs():
        # T4, L4, A10G, A100, H100
        concurrent_cn = int(os.getenv("IMAGE_CONCURRENT_CN", "1"))
        print(f"image_concurrent_cn:{concurrent_cn}")
        return concurrent_cn


# ----------------------- app -------------------------------
app = modal.App(ContainerRuntimeConfig.get_app_name())


# 128 MiB of memory and 0.125 CPU cores by default container runtime
@app.cls(
    image=ContainerRuntimeConfig.get_img(),
    gpu=ContainerRuntimeConfig.get_gpu(),
    secrets=[modal.Secret.from_name("achatbot")],
    cpu=2.0,
    container_idle_timeout=300,
    timeout=600,
    allow_concurrent_inputs=ContainerRuntimeConfig.get_allow_concurrent_inputs(),
)
class Srv:
    @modal.build()
    def download_model(self):
        # https://huggingface.co/docs/huggingface_hub/guides/download
        from huggingface_hub import snapshot_download
        from achatbot.common.types import MODELS_DIR

        os.makedirs(MODELS_DIR, exist_ok=True)
        print(f"start downloading model to dir:{MODELS_DIR}")

        # asr model repo
        if "sense_voice_asr" in os.getenv("ASR_TAG", "sense_voice_asr"):
            snapshot_download(
                repo_id="FunAudioLLM/SenseVoiceSmall",
                repo_type="model",
                allow_patterns="*",
                local_dir=os.path.join(MODELS_DIR, "FunAudioLLM/SenseVoiceSmall"),
            )

        # llm model repo
        llm_model_repo = os.getenv("LLM_MODEL_NAME_OR_PATH")
        if llm_model_repo:
            repo_id = "/".join(llm_model_repo.split("/")[-2:])
            snapshot_download(
                repo_id=repo_id,
                repo_type="model",
                allow_patterns="*",
                local_dir=os.path.join(MODELS_DIR, repo_id),
            )

        print("download model done")

    @modal.enter()
    def enter(self):
        print("start enter")

    @modal.asgi_app()
    def app(self):
        from achatbot.cmd.http.server.fastapi_daily_bot_serve import app

        return app


if __name__ == "__main__":
    pass



from types import MappingProxyType
import threading
import logging
import queue
import time
import io
import os

from pydub import AudioSegment

from achatbot.common.audio_stream.helper import AudioBufferManager
from achatbot.common.factory import EngineClass
from achatbot.common.interface import IAudioStream, IPlayer
from achatbot.common.session import Session
from achatbot.common.types import (
    AudioPlayerArgs,
    PYAUDIO_PAINT16,
    PYAUDIO_PAINT24,
    PYAUDIO_PAINT32,
    PYAUDIO_PAFLOAT32,
    PYAUDIO_PAINT8,
    PYAUDIO_PAUINT8,
    PYAUDIO_PACUSTOMFORMAT,
)
from achatbot.types.speech.audio_stream import AudioStreamInfo


class AudioPlayer(EngineClass):
    def __init__(self, **args) -> None:
        self.args = AudioPlayerArgs(**args)
        # self.audio = PyAudioStream(PyAudioStreamArgs(
        #    format=self.args.format,
        #    channels=self.args.channels,
        #    rate=self.args.rate,
        #    output_device_index=self.args.output_device_index,
        #    output=True,
        #    frames_per_buffer=self.args.frames_per_buffer,
        # ))

    def close(self):
        self.audio.close()

    def set_out_stream(self, audio_stream: EngineClass | IAudioStream):
        self.audio = audio_stream
        self.stream_info: AudioStreamInfo = audio_stream.get_stream_info()


class StreamPlayer(AudioPlayer, IPlayer):
    TAG = "stream_player"

    def __init__(self, **args) -> None:
        super().__init__(**args)

        self.first_chunk_played = False
        self.playback_active = False
        self.playback_thread: threading.Thread = None
        self.pause_event: threading.Event = threading.Event()
        self.immediate_stop_event: threading.Event = threading.Event()

        self.buffer_manager: AudioBufferManager = AudioBufferManager(queue.Queue())

    def open(self):
        self.audio.open_stream()

    def close(self):
        self.audio.close()
        self.playback_active = False

    def start(self, session: Session):
        self.first_chunk_played = False
        self.playback_active = True
        self.audio.start_stream()

        if not self.playback_thread or not self.playback_thread.is_alive():
            self.playback_thread = threading.Thread(target=self._process_buffer, args=(session,))
            self.playback_thread.start()

    def _process_buffer(self, session: Session):
        logging.info(f"{self.TAG} start process buffer thread")
        while self.playback_active or not self.buffer_manager.empty():
            try:
                if self.audio.is_stream_active() is False:
                    break
                chunk = self.buffer_manager.get_from_buffer(timeout=1.0)
                chunk and self._play_chunk(session, chunk)

                if self.immediate_stop_event.is_set():
                    logging.info(f"Immediate stop requested, aborting {self.TAG}")
                    break
            except queue.Empty:
                continue
            except Exception as ex:
                raise ex

        self.args.on_play_end and self.args.on_play_end(session)
        self.playback_active = False
        logging.info("stream player end")

    def _play_chunk(self, session: Session, chunk):
        # handle mpeg
        if self.stream_info.pyaudio_out_format == PYAUDIO_PACUSTOMFORMAT:
            # convert to pcm using pydub
            segment = AudioSegment.from_mp3(io.BytesIO(chunk))
            chunk = segment.raw_data

        sub_chunk_len = (
            self.stream_info.out_frames_per_buffer
            * self.stream_info.out_channels
            * self.stream_info.out_sample_width
        )
        for i in range(0, len(chunk), sub_chunk_len):
            sub_chunk = chunk[i : i + sub_chunk_len]

            if not self.first_chunk_played and self.args.on_play_start:
                self.on_play_start(session, sub_chunk)
                self.first_chunk_played = True

            if self.args.on_play_chunk:
                self.args.on_play_chunk(session, sub_chunk)

            self.audio.write_stream(sub_chunk)

            while self.pause_event.is_set():
                time.sleep(0.01)

            if self.immediate_stop_event.is_set():
                break

    def play_audio(self, session: Session):
        """
        send playing audio chunk to buffer (queue.Queue)
        """
        if session.ctx.state.get("tts_chunk") is None:
            return
        self.buffer_manager.add_to_buffer(session.ctx.state["tts_chunk"])

    def stop(self, session: Session):
        if not self.playback_thread:
            logging.warn("No playback thread found, cannot stop playback")
            return

        if self.args.is_immediate_stop:
            self.immediate_stop_event.set()
            # wait immediate stop play audio over
            while self.playback_active:
                time.sleep(0.1)
            self.immediate_stop_event.clear()
            return

        self.playback_active = False

        if self.playback_thread and self.playback_thread.is_alive():
            # wait current audio play over
            self.playback_thread.join()

        time.sleep(0.1)

        self.immediate_stop_event.clear()
        self.buffer_manager.clear_buffer()
        self.playback_thread = None
        self.audio.stop_stream()

    def pause(self):
        self.pause_event.set()

    def resume(self):
        self.pause_event.clear()


class PlayStreamInit:
    # TTS_TAG : stream_info, read_only dict
    map_tts_player_stream_info = MappingProxyType(
        {
            "tts_coqui": {
                "format": PYAUDIO_PAFLOAT32,
                "channels": 1,
                "rate": 24000,
                "sample_width": 4,
            },
            "tts_chat": {
                "format": PYAUDIO_PAINT16,
                "channels": 1,
                "rate": 24000,
                "sample_width": 2,
            },
            "tts_edge": {
                "format": PYAUDIO_PAINT16,
                "channels": 1,
                "rate": 22050,
                "sample_width": 2,
            },
            "tts_g": {
                "format": PYAUDIO_PAINT16,
                "channels": 1,
                "rate": 22050,
                "sample_width": 2,
            },
            "tts_cosy_voice": {
                "format": PYAUDIO_PAINT16,
                "channels": 1,
                "rate": 22050,
                "sample_width": 2,
            },
            "tts_cosy_voice2": {
                "format": PYAUDIO_PAINT16,
                "channels": 1,
                "rate": 22050,
                "sample_width": 2,
            },
            "tts_fishspeech": {
                "format": PYAUDIO_PAFLOAT32,
                "channels": 1,
                "rate": 44100,
                "sample_width": 2,
            },
            "tts_f5": {
                "format": PYAUDIO_PAFLOAT32,
                "channels": 1,
                "rate": 24000,
                "sample_width": 2,
            },
            "tts_openvoicev2": {
                "format": PYAUDIO_PAFLOAT32,
                "channels": 1,
                "rate": 22050,
                "sample_width": 2,
            },
            "tts_kokoro": {
                "format": PYAUDIO_PAFLOAT32,
                "channels": 1,
                "rate": 24000,
                "sample_width": 2,
            },
            "tts_onnx_kokoro": {
                "format": PYAUDIO_PAFLOAT32,
                "channels": 1,
                "rate": 24000,
                "sample_width": 2,
            },
            "tts_llasa": {
                "format": PYAUDIO_PAFLOAT32,
                "channels": 1,
                "rate": 16000,
                "sample_width": 2,
            },
            "tts_daily_speaker": {
                "format": PYAUDIO_PAINT16,
                "channels": 1,
                "rate": 16000,
                "sample_width": 2,
            },
            "tts_16k_speaker": {
                "format": PYAUDIO_PAINT16,
                "channels": 1,
                "rate": 16000,
                "sample_width": 2,
            },
        }
    )

    @staticmethod
    def get_stream_info() -> dict:
        tts_tag = os.getenv("TTS_TAG", "tts_edge")
        if tts_tag in PlayStreamInit.map_tts_player_stream_info:
            # !NOTE: return map_tts_player_stream_info is ref can change it,
            # so don't change it, just read only map (use MappingProxyType)
            return PlayStreamInit.map_tts_player_stream_info[tts_tag]
        return {}



import asyncio
import os
import logging

import unittest
from agora_realtime_ai_api import rtc
from agora_realtime_ai_api.token_builder import realtimekit_token_builder
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.frames.control_frames import EndFrame
from apipeline.processors.logger import FrameLogger
from apipeline.frames.data_frames import TextFrame, DataFrame, AudioRawFrame, ImageRawFrame

from src.services.help.agora.channel import AgoraChannel
from src.services.help import RoomManagerEnvInit
from src.common.types import AgoraParams
from src.common.logger import Logger
from src.transports.agora import AgoraTransport

from dotenv import load_dotenv

load_dotenv(override=True)

r"""
ROOM_NAME=bot-room \
    python -m unittest test.integration.processors.test_agora_echo_processor.TestProcessor

ROOM_NAME=bot-room RUN_TIMEOUT=3600 \
    python -m unittest test.integration.processors.test_agora_echo_processor.TestProcessor
"""


class TestProcessor(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        logger = Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False)
        cls.room_name = os.getenv("ROOM_NAME", "chat-room")
        rtc.logger = logger
        realtimekit_token_builder.logger = logger

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        self.agora_rtc_channel: AgoraChannel = RoomManagerEnvInit.initEngine("agora_rtc_channel")

    async def asyncTearDown(self):
        await self.agora_rtc_channel.close_session()

    async def test_echo(self):
        token = await self.agora_rtc_channel.gen_token(self.room_name)
        self.params = AgoraParams(
            audio_in_enabled=True,
            audio_in_sample_rate=16000,
            audio_in_channels=1,
            audio_out_enabled=True,
            audio_out_sample_rate=16000,
            audio_out_channels=1,
            camera_in_enabled=True,
            camera_out_enabled=True,
            camera_out_width=640,  # from Video Profiles https://webdemo.agora.io/basicVideoCall/index.html
            camera_out_height=480,
            camera_out_framerate=30,
        )
        transport = AgoraTransport(token, self.params)
        self.assertGreater(len(transport.event_names), 0)

        self.task = PipelineTask(
            Pipeline(
                [
                    transport.input_processor(),
                    # FrameLogger(include_frame_types=[AudioRawFrame]),
                    # FrameLogger(include_frame_types=[ImageRawFrame]),
                    transport.output_processor(),
                ]
            ),
            params=PipelineParams(allow_interruptions=False),
        )

        transport.add_event_handler("on_connected", self.on_connected)
        transport.add_event_handler("on_connection_state_changed", self.on_connection_state_changed)
        transport.add_event_handler("on_connection_failure", self.on_connection_failure)
        transport.add_event_handler("on_disconnected", self.on_disconnected)
        transport.add_event_handler("on_error", self.on_error)
        transport.add_event_handler("on_first_participant_joined", self.on_first_participant_joined)
        transport.add_event_handler("on_participant_disconnected", self.on_participant_disconnected)
        # sometime don't get on_connection_state_changed when disconnected
        transport.add_event_handler("on_connection_state_changed", self.on_connection_state_changed)
        transport.add_event_handler(
            "on_audio_subscribe_state_changed", self.on_audio_subscribe_state_changed
        )
        transport.add_event_handler("on_data_received", self.on_data_received)
        transport.add_event_handler(
            "on_video_subscribe_state_changed", self.on_video_subscribe_state_changed
        )

        runner = PipelineRunner()
        try:
            await asyncio.wait_for(runner.run(self.task), int(os.getenv("RUN_TIMEOUT", "60")))
        except asyncio.TimeoutError:
            logging.warning("Test run timeout. Exiting")
            await self.task.queue_frame(EndFrame())

    async def on_connected(
        self,
        transport: AgoraTransport,
        agora_rtc_conn: rtc.RTCConnection,
        conn_info: rtc.RTCConnInfo,
        reason: int,
    ):
        logging.info(f"agora_rtc_conn:{agora_rtc_conn} conn_info:{conn_info} reason:{reason}")

    async def on_connection_failure(self, transport: AgoraTransport, reason: int):
        logging.info(f"reason:{reason}")

    async def on_error(self, transport: AgoraTransport, error_msg: str):
        logging.info(f"error_msg:{error_msg}")

    async def on_first_participant_joined(self, transport: AgoraTransport, user_id: str):
        logging.info(f"user_id:{user_id}")

        if self.params.audio_in_enabled:
            transport.capture_participant_audio(user_id)
        if self.params.camera_in_enabled:
            # passive capture one image
            # transport.capture_participant_video(user_id, framerate=0)
            # active to capture
            transport.capture_participant_video(user_id)

        await transport.send_message(
            "hello,你好，我是机器人。",
            participant_id=user_id,
        )

    async def on_participant_disconnected(
        self,
        transport: AgoraTransport,
        agora_rtc_conn: rtc.RTCConnection,
        user_id: str,
        reason: int,
    ):
        logging.info(f"Partcipant {user_id} left. reason:{reason}")
        logging.info(f"current remote Partcipants {transport.get_participant_ids()}")

    async def on_disconnected(
        self,
        transport: AgoraTransport,
        agora_rtc_conn: rtc.RTCConnection,
        conn_info: rtc.RTCConnInfo,
        reason: int,
    ):
        logging.info("disconnected reason %s, Exiting." % reason)
        await self.task.queue_frame(EndFrame())

    async def on_connection_state_changed(
        self,
        transport: AgoraTransport,
        agora_rtc_conn: rtc.RTCConnection,
        conn_info: rtc.RTCConnInfo,
        reason: int,
    ):
        logging.info(f"connection state {conn_info.state} reason:{reason}")

    async def on_audio_subscribe_state_changed(
        self,
        transport: AgoraTransport,
        agora_local_user: rtc.LocalUser,
        channel: str,
        user_id: str,
        old_state: int,
        new_state: int,
        elapse_since_last_state: int,
    ):
        logging.info(
            f"agora_local_user:{agora_local_user}, channel:{channel}, user_id:{user_id}, old_state:{old_state}, new_state:{new_state}, elapse_since_last_state:{elapse_since_last_state}"
        )

    async def on_data_received(self, transport: AgoraTransport, data: bytes, user_id: str):
        logging.info(f"len(data):{len(data)} user_id:{user_id}")

    async def on_video_subscribe_state_changed(
        self,
        transport: AgoraTransport,
        agora_local_user: rtc.LocalUser,
        channel: str,
        user_id: str,
        old_state: int,
        new_state: int,
        elapse_since_last_state: int,
    ):
        logging.info(
            f"agora_local_user:{agora_local_user}, channel:{channel}, user_id:{user_id}, old_state:{old_state}, new_state:{new_state}, elapse_since_last_state:{elapse_since_last_state}"
        )



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
# Copyright FunASR (https://github.com/FunAudioLLM/SenseVoice). All Rights Reserved.
#  MIT License  (https://opensource.org/licenses/MIT)

from funasr import AutoModel
from funasr.utils.postprocess_utils import rich_transcription_postprocess

model_dir = "iic/SenseVoiceSmall"


model = AutoModel(
    model=model_dir,
    trust_remote_code=True,
    remote_code="./model.py",
    vad_model="fsmn-vad",
    vad_kwargs={"max_single_segment_time": 30000},
    device="cuda:0",
)

# en
res = model.generate(
    input=f"{model.model_path}/example/en.mp3",
    cache={},
    language="auto",  # "zh", "en", "yue", "ja", "ko", "nospeech"
    use_itn=True,
    batch_size_s=60,
    merge_vad=True,  #
    merge_length_s=15,
)
text = rich_transcription_postprocess(res[0]["text"])
print(text)

# zh
res = model.generate(
    input=f"{model.model_path}/example/zh.mp3",
    cache={},
    language="auto",  # "zh", "en", "yue", "ja", "ko", "nospeech"
    use_itn=True,
    batch_size_s=60,
    merge_vad=True,  #
    merge_length_s=15,
)
text = rich_transcription_postprocess(res[0]["text"])
print(text)

# yue
res = model.generate(
    input=f"{model.model_path}/example/yue.mp3",
    cache={},
    language="auto",  # "zh", "en", "yue", "ja", "ko", "nospeech"
    use_itn=True,
    batch_size_s=60,
    merge_vad=True,  #
    merge_length_s=15,
)
text = rich_transcription_postprocess(res[0]["text"])
print(text)

# ja
res = model.generate(
    input=f"{model.model_path}/example/ja.mp3",
    cache={},
    language="auto",  # "zh", "en", "yue", "ja", "ko", "nospeech"
    use_itn=True,
    batch_size_s=60,
    merge_vad=True,  #
    merge_length_s=15,
)
text = rich_transcription_postprocess(res[0]["text"])
print(text)


# ko
res = model.generate(
    input=f"{model.model_path}/example/ko.mp3",
    cache={},
    language="auto",  # "zh", "en", "yue", "ja", "ko", "nospeech"
    use_itn=True,
    batch_size_s=60,
    merge_vad=True,  #
    merge_length_s=15,
)
text = rich_transcription_postprocess(res[0]["text"])
print(text)



from dataclasses import dataclass, field

import torch


@dataclass
class TransformersLMArgs:
    r"""
    HF transformers language model args (tiny,small,large,huge)
    """

    lm_model_name_or_path: str = field(
        default="HuggingFaceTB/SmolLM-360M-Instruct",
        metadata={
            "help": "The pretrained language model to use. Default is 'HuggingFaceTB/SmolLM-360M-Instruct'."
        },
    )
    lm_device_map: str | dict | None = field(
        default=None,
        metadata={
            "help": "The device map for multi cpu/gpu. use 'cuda' for GPU acceleration, 'mps','cpu','auto', dict, None. default None"
        },
    )
    lm_device: str = field(
        default="cpu",
        metadata={
            "help": "The device for single cpu/mps/gpu. use 'cuda' for GPU acceleration, 'mps'(apple), 'cpu'. default cuda"
        },
    )
    lm_torch_dtype: str = field(
        default="auto",
        metadata={
            "help": "The PyTorch data type for the model and input tensors. One of `float32` (full-precision), `float16` or `bfloat16` (both half-precision), auto. default auto."
        },
    )
    # https://huggingface.co/docs/transformers/perf_infer_gpu_one
    lm_attn_impl: str | None = field(
        default=None,
        metadata={
            "help": "The attention implementation to use. One of 'sdpa', 'flash_attention_2', default no attention implementation."
        },
    )
    user_role: str = field(
        default="user",
        metadata={"help": "Role assigned to the user in the chat context. Default is 'user'."},
    )
    warnup_prompt: str = field(
        default="Repeat the word 'weedge niu bi'.",
        metadata={"help": "warnup llm generate prompt. Default is 'weedge niu bi'."},
    )
    warnup_steps: int = field(
        default=2,
        metadata={"help": "The number of steps to run the warmup prompt. Default is 2."},
    )
    init_chat_role: str = field(
        default="system",
        metadata={"help": "Initial role for setting up the chat context. Default is 'system'."},
    )
    init_chat_prompt: str = field(
        default="You are a helpful and friendly AI assistant. You are polite, respectful, and aim to provide concise responses of less than 20 words.",
        metadata={
            "help": "The initial chat prompt to establish context for the language model. Default is 'You are a helpful AI assistant.'"
        },
    )
    lm_gen_seed: int = field(
        default=42,
        metadata={"help": "The seed to use for the language model. Default is 42."},
    )
    lm_gen_max_new_tokens: int = field(
        default=1024,
        metadata={
            "help": "Maximum number of new tokens to generate in a single completion. Default is 1024."
        },
    )
    lm_gen_min_new_tokens: int = field(
        default=0,
        metadata={
            "help": "Minimum number of new tokens to generate in a single completion. Default is 0."
        },
    )
    lm_gen_do_sample: bool = field(
        default=True,
        metadata={
            "help": "Whether to use sampling; set this to False for deterministic outputs. Default is False."
        },
    )
    lm_gen_temperature: float = field(
        default=0.1,
        metadata={
            "help": "Controls the randomness of the output. Set to 0.0 for deterministic (repeatable) outputs. Default is 0.1."
        },
    )
    lm_gen_top_k: int = field(
        default=1,
        metadata={
            "help": "Changing the top - k parameter sets the size of the shortlist the model samples from as it outputs each token. Setting top - k to 1 gives us greedy decoding. Default is 1"
        },
    )
    lm_gen_top_p: float = field(
        default=0.8,
        metadata={
            "help": "Top-p is usually set to a high value (like 0.75) with the purpose of limiting the long tail of low-probability tokens that may be sampled. We can use both top-k and top-p together. If both k and p are enabled, p acts after k. Default is 0.8."
        },
    )
    lm_gen_repetition_penalty: float = field(
        default=1.0,
        metadata={
            "help": "Controls the token repetition pealty.  no repetition Default is 1.0, >1.0: low repetition, <1.0: high repetition"
        },
    )
    lm_tokenizer_decode_batch_size: int = field(
        default=60,
        metadata={"help": "Number of tokens to batch decode at a time"},
    )
    chat_history_size: int | None = field(
        default=None,
        metadata={
            "help": "Number of interactions assitant-user to keep for the chat. None for no limitations, <= 0 for no history"
        },
    )
    lm_stream: bool = field(
        default=True,
        metadata={
            "help": "Whether to use streaming; set this to True for streaming output. Default is True."
        },
    )
    model_type: str = field(
        default="chat_completion",
        metadata={
            "help": "Model type, generate, chat_completion(chat/instruct). Default is chat_completion"
        },
    )
    lm_bnb_quant_type: str = field(
        default="int4",
        metadata={"help": "The BitsAndBytes quantization type, default int4."},
    )

    def to_dict(self) -> dict:
        return {
            field.name: getattr(self, field.name) for field in self.__dataclass_fields__.values()
        }


@dataclass
class TransformersSpeechLMArgs(TransformersLMArgs):
    """
    text+speech lm args
    """

    def to_dict(self) -> dict:
        return super().to_dict()


@dataclass
class TransformersImageLMArgs(TransformersLMArgs):
    """
    text+Image lm args
    """

    vae_model_name_or_path: str = field(
        default="stabilityai/sdxl-vae",
        metadata={"help": "The diffusion model to use. Default is 'stabilityai/sdxl-vae'."},
    )

    def to_dict(self) -> dict:
        return super().to_dict()



import requests
import json
import os

import logging

from src.common.types import SerperApiArgs
from .api import SearchBaseApi

from dotenv import load_dotenv

load_dotenv(override=True)


class SerperApi(SearchBaseApi):
    BASE_URL = "https://google.serper.dev/search"
    TAG = "serper_api"

    def __init__(self, **args) -> None:
        super().__init__()
        self.args = SerperApiArgs(**args)

    def _web_search(self, session, query: str) -> str:
        api_key = os.getenv("SERPER_API_KEY", "")
        url = self.BASE_URL
        headers = {
            "X-API-KEY": api_key,
            "Content-Type": "application/json",
        }
        payload = {
            "q": query,
            "gl": self.args.gl,
            "hl": self.args.hl,
            "page": self.args.page,
            "num": self.args.num,
        }

        try:
            response = self.requests.post(url, headers=headers, json=payload)
            response.raise_for_status()  # Raises for HTTP errors
            data = response.json()
            snippets = [item["snippet"] for item in data["organic"]]
            return json.dumps(snippets)
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
            return json.dumps({"error": "Failed to fetch search results"})
        except Exception as err:
            print(f"An error occurred: {err}")
            return json.dumps({"error": "Failed to fetch search results"})



import asyncio
import logging
import os
import argparse

from fastapi import WebSocket
from dotenv import load_dotenv

from src.cmd.bots.bot_loader import BotLoader
from src.cmd.bots.base_fastapi_websocket_server import AIFastapiWebsocketBot
from src.common.types import CONFIG_DIR
from src.common.const import *
from src.common.logger import Logger
from src.cmd.http.server.fastapi_daily_bot_serve import app, ngrok_proxy


load_dotenv(override=True)
Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)


@app.websocket("/")
async def websocket_endpoint(websocket: WebSocket):
    run_bot: AIFastapiWebsocketBot = await BotLoader.load_bot(config.f, bot_type="fastapi_ws_bot")

    # NOTE: after init, websocket to accept connection, then to run
    await websocket.accept()
    run_bot.set_fastapi_websocket(websocket)
    logging.info(f"accept client: {websocket.client}")
    await run_bot.try_run()


if __name__ == "__main__":
    import uvicorn

    default_host = os.getenv("HOST", "0.0.0.0")
    default_port = int(os.getenv("FAST_API_PORT", "4321"))

    parser = argparse.ArgumentParser(description="Fastapi Websocket Bot Runner")
    parser.add_argument("--host", type=str, default=default_host, help="Host address")
    parser.add_argument("--port", type=int, default=default_port, help="Port number")
    parser.add_argument("--reload", action="store_true", help="Reload code on change")
    parser.add_argument("--ngrok", action="store_true", help="use ngrok proxy")
    parser.add_argument(
        "-f",
        type=str,
        default=os.path.join(CONFIG_DIR, "bots/dummy_bot.json"),
        help="Bot configuration json file",
    )

    config = parser.parse_args()

    if config.ngrok:
        ngrok_proxy(config.port)

    # Note: not event loop to new one
    # run_bot: AIFastapiWebsocketBot = asyncio.get_event_loop().run_until_complete(BotLoader.load_bot(config.f, bot_type="fastapi_ws_bot"))

    # use one event loop to run
    # run_bot: AIFastapiWebsocketBot = asyncio.run(
    #    BotLoader.load_bot(config.f, bot_type="fastapi_ws_bot"))

    # api docs: http://0.0.0.0:4321/docs
    uvicorn.run(
        "src.cmd.http.server.fastapi_daily_bot_serve:app",
        host=config.host,
        port=config.port,
        reload=config.reload,
    )



import asyncio
import logging
import os
from signal import SIGINT, SIGTERM

import cv2
import mediapipe as mp
import numpy as np
from mediapipe import solutions
from mediapipe.framework.formats import landmark_pb2

from livekit import api, rtc

# ensure LIVEKIT_URL, LIVEKIT_API_KEY, and LIVEKIT_API_SECRET are set

tasks = set()

# You can download a face landmark model file from
# https://developers.google.com/mediapipe/solutions/vision/face_landmarker#models
model_path = "./models/face_landmarker.task"

BaseOptions = mp.tasks.BaseOptions
FaceLandmarker = mp.tasks.vision.FaceLandmarker
FaceLandmarkerOptions = mp.tasks.vision.FaceLandmarkerOptions
VisionRunningMode = mp.tasks.vision.RunningMode

options = FaceLandmarkerOptions(
    base_options=BaseOptions(model_asset_path=model_path),
    running_mode=VisionRunningMode.VIDEO,
)


async def main(room: rtc.Room) -> None:
    video_stream = None

    @room.on("track_subscribed")
    def on_track_subscribed(track: rtc.Track, *_):
        if track.kind == rtc.TrackKind.KIND_VIDEO:
            nonlocal video_stream
            if video_stream is not None:
                # only process the first stream received
                return

            print("subscribed to track: " + track.name)
            video_stream = rtc.VideoStream(track, format=rtc.VideoBufferType.RGB24)
            task = asyncio.create_task(frame_loop(video_stream))
            tasks.add(task)
            task.add_done_callback(tasks.remove)

    token = (
        api.AccessToken()
        .with_identity("face-landmark-bot")
        .with_name("face landmark Bot")
        .with_grants(
            api.VideoGrants(
                room_join=True,
                room="chat-room",
            )
        )
    )
    await room.connect(os.getenv("LIVEKIT_URL"), token.to_jwt())
    print("connected to room: " + room.name)


def draw_landmarks_on_image(rgb_image, detection_result):
    # from
    # https://github.com/googlesamples/mediapipe/blob/main/examples/face_landmarker/python/%5BMediaPipe_Python_Tasks%5D_Face_Landmarker.ipynb
    face_landmarks_list = detection_result.face_landmarks

    # Loop through the detected faces to visualize.
    for face_landmarks in face_landmarks_list:
        # Draw the face landmarks.
        face_landmarks_proto = landmark_pb2.NormalizedLandmarkList()
        face_landmarks_proto.landmark.extend(
            [
                landmark_pb2.NormalizedLandmark(x=landmark.x, y=landmark.y, z=landmark.z)
                for landmark in face_landmarks
            ]
        )

        solutions.drawing_utils.draw_landmarks(
            image=rgb_image,
            landmark_list=face_landmarks_proto,
            connections=mp.solutions.face_mesh.FACEMESH_TESSELATION,
            landmark_drawing_spec=None,
            connection_drawing_spec=mp.solutions.drawing_styles.get_default_face_mesh_tesselation_style(),
        )
        solutions.drawing_utils.draw_landmarks(
            image=rgb_image,
            landmark_list=face_landmarks_proto,
            connections=mp.solutions.face_mesh.FACEMESH_CONTOURS,
            landmark_drawing_spec=None,
            connection_drawing_spec=mp.solutions.drawing_styles.get_default_face_mesh_contours_style(),
        )
        solutions.drawing_utils.draw_landmarks(
            image=rgb_image,
            landmark_list=face_landmarks_proto,
            connections=mp.solutions.face_mesh.FACEMESH_IRISES,
            landmark_drawing_spec=None,
            connection_drawing_spec=mp.solutions.drawing_styles.get_default_face_mesh_iris_connections_style(),
        )


async def frame_loop(video_stream: rtc.VideoStream) -> None:
    landmarker = FaceLandmarker.create_from_options(options)
    cv2.namedWindow("livekit_video", cv2.WINDOW_AUTOSIZE)
    cv2.startWindowThread()
    async for frame_event in video_stream:
        print(frame_event)
        buffer = frame_event.frame
        print(buffer.type, len(buffer.data))

        arr = np.frombuffer(buffer.data, dtype=np.uint8)
        arr = arr.reshape((buffer.height, buffer.width, 3))

        mp_image = mp.Image(image_format=mp.ImageFormat.SRGB, data=arr)
        detection_result = landmarker.detect_for_video(mp_image, frame_event.timestamp_us)

        draw_landmarks_on_image(arr, detection_result)

        arr = cv2.cvtColor(arr, cv2.COLOR_RGB2BGR)
        cv2.imshow("livekit_video", arr)
        if cv2.waitKey(1) & 0xFF == ord("q"):
            break

    landmarker.close()
    cv2.destroyAllWindows()


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        handlers=[logging.FileHandler("face_landmark.log"), logging.StreamHandler()],
    )

    loop = asyncio.get_event_loop()
    room = rtc.Room(loop=loop)

    async def cleanup():
        await room.disconnect()
        loop.stop()

    asyncio.ensure_future(main(room))
    for signal in [SIGINT, SIGTERM]:
        loop.add_signal_handler(signal, lambda: asyncio.ensure_future(cleanup()))

    try:
        loop.run_forever()
    finally:
        loop.close()



import logging
import os

from src.common.types import DEFAULT_SYSTEM_PROMPT, MODELS_DIR, PersonalAIProxyArgs
from src.types.llm.transformers import TransformersLMArgs
from src.common import interface
from src.common.factory import EngineClass, EngineFactory

from dotenv import load_dotenv

load_dotenv(override=True)


class LLMEnvInit:
    @staticmethod
    def getEngine(tag, **kwargs) -> interface.IAsr | EngineClass:
        if "llm_llamacpp" in tag:
            from . import llamacpp
        elif "llm_personalai_proxy" in tag:
            from . import personalai
        elif "llm_transformers_manual_vision_qwen" in tag:
            from .transformers import manual_vision_qwen
        elif "llm_transformers_manual_vision_llama" in tag:
            from .transformers import manual_vision_llama
        elif "llm_transformers_manual_vision_molmo" in tag:
            from .transformers import manual_vision_molmo
        elif "llm_transformers_manual_vision_janus_flow" in tag:
            from .transformers import manual_vision_img_janus_flow
        elif "llm_transformers_manual_image_janus_flow" in tag:
            from .transformers import manual_vision_img_janus_flow
        elif "llm_transformers_manual_vision_janus" in tag:
            from .transformers import manual_vision_img_janus_pro
        elif "llm_transformers_manual_image_janus" in tag:
            from .transformers import manual_vision_img_janus_pro
        elif "llm_transformers_manual" in tag:
            from .transformers import manual
        elif "llm_transformers_pipeline" in tag:
            from .transformers import pipeline

        engine = EngineFactory.get_engine_by_tag(EngineClass, tag, **kwargs)
        return engine

    @staticmethod
    def initLLMEngine(
        tag: str | None = None, kwargs: dict | None = None
    ) -> interface.ILlm | EngineClass:
        # llm
        tag = tag if tag else os.getenv("LLM_TAG", "llm_llamacpp")
        if kwargs is None:
            kwargs = LLMEnvInit.map_config_func[tag]()
        engine = LLMEnvInit.getEngine(tag, **kwargs)
        logging.info(f"initLLMEngine: {tag}, {engine}")
        return engine

    @staticmethod
    def get_llm_llamacpp_args() -> dict:
        kwargs = {}
        kwargs["model_name"] = os.getenv("LLM_MODEL_NAME", "qwen2")
        kwargs["model_path"] = os.getenv(
            "LLM_MODEL_PATH", os.path.join(MODELS_DIR, "qwen2-1_5b-instruct-q8_0.gguf")
        )
        kwargs["model_type"] = os.getenv("LLM_MODEL_TYPE", "chat")
        kwargs["n_threads"] = os.cpu_count()
        kwargs["verbose"] = False
        kwargs["n_gpu_layers"] = int(os.getenv("N_GPU_LAYERS", "0"))
        kwargs["flash_attn"] = bool(os.getenv("FLASH_ATTN", ""))
        kwargs["llm_stream"] = True
        # if logger.getEffectiveLevel() != logging.DEBUG:
        #    kwargs["verbose"] = False
        kwargs["llm_stop"] = [
            "<|end|>",
            "<|im_end|>",
            "<|endoftext|>",
            "<{/end}>",
            "</s>",
            "/s>",
            "</s",
            "<s>",
            "<|user|>",
            "<|assistant|>",
            "<|system|>",
        ]
        kwargs["llm_chat_system"] = os.getenv("LLM_CHAT_SYSTEM", DEFAULT_SYSTEM_PROMPT)
        kwargs["llm_tool_choice"] = os.getenv("LLM_TOOL_CHOICE", None)
        kwargs["chat_format"] = os.getenv("LLM_CHAT_FORMAT", None)
        kwargs["tokenizer_path"] = os.getenv("LLM_TOKENIZER_PATH", None)
        kwargs["clip_model_path"] = os.getenv("LLM_CLIP_MODEL_PATH", None)
        return kwargs

    @staticmethod
    def get_llm_personal_ai_proxy_args() -> dict:
        kwargs = PersonalAIProxyArgs(
            api_url=os.getenv("API_URL", "http://localhost:8787/"),
            chat_bot=os.getenv("CHAT_BOT", "openai"),
            model_type=os.getenv("CHAT_TYPE", "chat_only"),
            openai_api_base_url=os.getenv("OPENAI_API_BASE_URL", "https://api.groq.com/openai/v1/"),
            model_name=os.getenv("LLM_MODEL_NAME", "llama3-70b-8192"),
            llm_chat_system=os.getenv("LLM_CHAT_SYSTEM", DEFAULT_SYSTEM_PROMPT),
            llm_stream=False,
            llm_max_tokens=int(os.getenv("LLM_MAX_TOKENS", "1024")),
            func_search_name=os.getenv("FUNC_SEARCH_NAME", "search_api"),
            func_weather_name=os.getenv("FUNC_WEATHER_NAME", "openweathermap_api"),
        ).__dict__
        return kwargs

    @staticmethod
    def _get_llm_transformers_args() -> dict:
        kwargs = TransformersLMArgs(
            lm_model_name_or_path=os.getenv(
                "LLM_MODEL_NAME_OR_PATH", os.path.join(MODELS_DIR, "Qwen/Qwen2-0.5B-Instruct")
            ),
            lm_attn_impl=os.getenv("LLM_ATTN_IMPL", None),
            lm_device=os.getenv("LLM_DEVICE", "cpu"),
            lm_torch_dtype=os.getenv("LLM_TORCH_DTYPE", "auto"),
            lm_gen_max_new_tokens=int(os.getenv("LLM_GEN_MAX_NEW_TOKENS", "1024")),
            lm_gen_min_new_tokens=int(os.getenv("LLM_GEN_MIN_NEW_TOKENS", "0")),
            lm_gen_do_sample=bool(os.getenv("LLM_GEN_DO_SAMPLE", "")),
            lm_gen_temperature=float(os.getenv("LLM_GEN_TEMPERATURE", "0.0")),
            lm_stream=bool(os.getenv("LLM_STREAM", "1")),
            init_chat_prompt=os.getenv("LLM_INIT_CHAT_PROMPT", ""),
            chat_history_size=int(os.getenv("LLM_CHAT_HISTORY_SIZE", "10")),  # cache 10 round
            model_type=os.getenv("LLM_MODEL_TYPE", "chat_completion"),
        ).__dict__
        return kwargs

    @staticmethod
    def get_llm_transformers_args() -> dict:
        hf_parser_file = os.getenv("HF_TRANSFOMERS_PARSER_FILE", None)
        if hf_parser_file:
            from .transformers import TransformersLLMInit

            return TransformersLLMInit.get_transformers_llm_args(hf_parser_file)
        return LLMEnvInit._get_llm_transformers_args()

    @staticmethod
    def get_llm_transformers_manual_image_janus_flow_args() -> dict:
        kwargs = LLMEnvInit.get_llm_transformers_args()
        kwargs["vae_model_name_or_path"] = os.getenv(
            "VAE_MODEL_NAME_OR_PATH",
            os.path.join(MODELS_DIR, "stabilityai/sdxl-vae"),
        )
        return kwargs

    # TAG : config
    map_config_func = {
        "llm_llamacpp": get_llm_llamacpp_args,
        "llm_personalai_proxy": get_llm_personal_ai_proxy_args,
        "llm_transformers_manual": get_llm_transformers_args,
        "llm_transformers_pipeline": get_llm_transformers_args,
        "llm_transformers_manual_vision_qwen": get_llm_transformers_args,
        "llm_transformers_manual_vision_llama": get_llm_transformers_args,
        "llm_transformers_manual_vision_molmo": get_llm_transformers_args,
        "llm_transformers_manual_vision_janus": get_llm_transformers_args,
        "llm_transformers_manual_image_janus": get_llm_transformers_args,
        "llm_transformers_manual_vision_janus_flow": get_llm_transformers_args,
        "llm_transformers_manual_image_janus_flow": get_llm_transformers_manual_image_janus_flow_args,
    }



import os
import logging

import unittest

from src.common.utils.img_utils import image_to_base64_data_uri
from src.common.interface import ILlm
from src.common.factory import EngineFactory, EngineClass
from src.core.llm.llamacpp import LLamacppLLM
from src.common.logger import Logger
from src.common.session import Session
from src.common.types import SessionCtx, MODELS_DIR, TEST_DIR
from src.cmd.init import PromptInit
from src.core.llm import LLMEnvInit

from dotenv import load_dotenv

load_dotenv(override=True)


r"""
python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_have_special_char

MODEL_TYPE=generate  python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_generate
MODEL_TYPE=generate STREAM=1 python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_generate

MODEL_TYPE=chat CHAT_FORMAT=chatml python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion
MODEL_TYPE=chat CHAT_FORMAT=chatml STREAM=1 python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion

MODEL_TYPE=chat-func \
    MODEL_NAME="functionary" \
    MODEL_PATH="./models/meetkai/functionary-small-v2.4-GGUF/functionary-small-v2.4.Q4_0.gguf" \
    TOKENIZER_PATH="./models/meetkai/functionary-small-v2.4-GGUF" \
    CHAT_FORMAT=functionary-v2 \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion

MODEL_TYPE=chat-func STREAM=1 \
    MODEL_NAME="functionary" \
    MODEL_PATH="./models/meetkai/functionary-small-v2.4-GGUF/functionary-small-v2.4.Q4_0.gguf" \
    TOKENIZER_PATH="./models/meetkai/functionary-small-v2.4-GGUF" \
    CHAT_FORMAT=functionary-v2 \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion

MODEL_TYPE=chat-func \
    MODEL_NAME="functionary" \
    MODEL_PATH="./models/meetkai/functionary-small-v2.4-GGUF/functionary-small-v2.4.Q4_0.gguf" \
    TOKENIZER_PATH="./models/meetkai/functionary-small-v2.4-GGUF" \
    CHAT_FORMAT=functionary-v2 \
    LLM_TOOL_CHOICE=auto \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion

MODEL_TYPE=chat-func PROMPT="查询下今日美股股价行情" \
    MODEL_NAME="functionary" \
    MODEL_PATH="./models/meetkai/functionary-small-v2.4-GGUF/functionary-small-v2.4.Q4_0.gguf" \
    TOKENIZER_PATH="./models/meetkai/functionary-small-v2.4-GGUF" \
    CHAT_FORMAT=functionary-v2 \
    LLM_TOOL_CHOICE=auto \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion

MODEL_TYPE=chat-func STREAM=1 PROMPT="查询下今日美股股价行情" \
    MODEL_NAME="functionary" \
    MODEL_PATH="./models/meetkai/functionary-small-v2.4-GGUF/functionary-small-v2.4.Q4_0.gguf" \
    TOKENIZER_PATH="./models/meetkai/functionary-small-v2.4-GGUF" \
    CHAT_FORMAT=functionary-v2 \
    LLM_TOOL_CHOICE=auto \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion

MODEL_TYPE=chat STREAM=1 \
    MODEL_NAME="minicpm-v-2.6" \
    MODEL_PATH="./models/openbmb/MiniCPM-V-2_6-gguf/ggml-model-Q4_0.gguf" \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_completion

MODEL_TYPE=chat STREAM=1 PROMPT="请描述一下图片内容" \
    MODEL_NAME="minicpm-v-2.6" \
    MODEL_PATH="./models/openbmb/MiniCPM-V-2_6-gguf/ggml-model-Q4_0.gguf" \
    CLIP_MODEL_PATH="./models/openbmb/MiniCPM-V-2_6-gguf/mmproj-model-f16.gguf" \
    CHAT_FORMAT=minicpm-v-2.6 \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_vision_img

MODEL_TYPE=chat STREAM=1 PROMPT="请描述一下图片内容" \
    MODEL_NAME="minicpm-v-2.6" \
    MODEL_PATH="./models/openbmb/MiniCPM-V-2_6-gguf/ggml-model-Q4_0.gguf" \
    CLIP_MODEL_PATH="./models/openbmb/MiniCPM-V-2_6-gguf/mmproj-model-f16.gguf" \
    CHAT_FORMAT=minicpm-v-2.6 \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_vision_local_img

MODEL_TYPE=chat STREAM=1 PROMPT="请描述一下图片内容" \
    LLM_MODEL_NAME="minicpm-v-2.6" \
    LLM_MODEL_PATH="./models/openbmb/MiniCPM-V-2_6-gguf/ggml-model-Q4_0.gguf" \
    LLM_CLIP_MODEL_PATH="./models/openbmb/MiniCPM-V-2_6-gguf/mmproj-model-f16.gguf" \
    LLM_CHAT_FORMAT=minicpm-v-2.6 \
    python -m unittest test.core.llm.test_llamacpp.TestLLamacppLLM.test_chat_vision_local_img_env_init
"""


class TestLLamacppLLM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.llm_tag = os.getenv("LLM_TAG", "llm_llamacpp")
        cls.prompt = os.getenv("PROMPT", "你好")
        cls.stream = os.getenv("STREAM", "")
        cls.model_type = os.getenv("MODEL_TYPE", "generate")
        cls.chat_format = os.getenv("CHAT_FORMAT", None)
        cls.llm_tool_choice = os.getenv("LLM_TOOL_CHOICE", None)
        cls.model_name = os.getenv("MODEL_NAME", "qwen2")
        cls.model_path = os.getenv(
            "MODEL_PATH", os.path.join(MODELS_DIR, "qwen2-1_5b-instruct-q8_0.gguf")
        )
        cls.tokenizer_path = os.getenv("TOKENIZER_PATH", None)
        cls.clip_model_path = os.getenv("CLIP_MODEL_PATH", None)
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        kwargs = {}
        kwargs["verbose"] = True
        kwargs["model_name"] = self.model_name
        kwargs["model_path"] = self.model_path
        kwargs["tokenizer_path"] = self.tokenizer_path
        kwargs["model_type"] = self.model_type
        kwargs["chat_format"] = self.chat_format
        kwargs["llm_tool_choice"] = self.llm_tool_choice
        kwargs["n_threads"] = os.cpu_count()
        kwargs["n_gpu_layers"] = int(os.getenv("N_GPU_LAYERS", "0"))
        kwargs["flash_attn"] = bool(os.getenv("FLASH_ATTN", ""))
        kwargs["llm_stop"] = ["<|end|>", "<|im_end|>", "<s>", "</s>"]
        kwargs["clip_model_path"] = self.clip_model_path
        self.kwargs = kwargs
        self.session = Session(**SessionCtx("test_client_id").__dict__)

    def tearDown(self):
        pass

    def test_generate(self):
        self.llm: LLamacppLLM = EngineFactory.get_engine_by_tag(
            EngineClass, self.llm_tag, **self.kwargs
        )
        self.llm.args.llm_stream = bool(self.stream)
        logging.debug(self.llm.args)
        self.assertEqual(self.llm.model_name(), self.llm.args.model_name)
        self.session.ctx.state["prompt"] = self.prompt
        logging.debug(self.session.ctx.state)
        iter = self.llm.generate(self.session)
        for item in iter:
            print(f"item--> {item}")
            self.assertGreater(len(item), 0)

    def test_have_special_char(self):
        self.llm: LLamacppLLM = EngineFactory.get_engine_by_tag(
            EngineClass, self.llm_tag, **self.kwargs
        )
        index = self.llm._have_special_char("""'你好',中国""")
        print(index)
        self.assertGreater(index, -1)

    def test_generate_with_system(self):
        self.llm: LLamacppLLM = EngineFactory.get_engine_by_tag(
            EngineClass, self.llm_tag, **self.kwargs
        )
        self.llm.args.llm_stream = bool(self.stream)
        logging.debug(self.llm.args)
        self.assertEqual(self.llm.model_name(), self.llm.args.model_name)
        history = []
        history.append(PromptInit.get_user_prompt(self.llm.model_name(), self.prompt))
        self.session.ctx.state["prompt"] = PromptInit.create_prompt(
            self.llm.model_name(),
            history,
            system_prompt="你是一个智能助手，回答请限制在1-5句话内。",
        )
        logging.debug(self.session.ctx.state)
        iter = self.llm.generate(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)

    def test_chat_completion(self):
        self.llm: LLamacppLLM = EngineFactory.get_engine_by_tag(
            EngineClass, self.llm_tag, **self.kwargs
        )
        self.llm.args.llm_stream = bool(self.stream)
        self.llm.args.llm_chat_system = "你是一个中国人,请用中文回答。回答限制在1-5句话内。要友好、乐于助人且简明扼要。保持对话简短而甜蜜。只用纯文本回答，不要包含链接或其他附加内容。不要回复计算机代码以及数学公式。"
        self.session.ctx.state["prompt"] = self.prompt
        logging.debug(self.session.ctx)
        iter = self.llm.chat_completion(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)

    def test_generate_case(self):
        os.environ["LLM_TAG"] = "llm_llamacpp"
        engine: LLamacppLLM = LLMEnvInit.initLLMEngine()
        engine.args.llm_chat_system = "你是一个中国人,请用中文回答。回答限制在1-5句话内。要友好、乐于助人且简明扼要。保持对话简短而甜蜜。只用纯文本回答，不要包含链接或其他附加内容。不要回复计算机代码以及数学公式。"
        self.session.ctx.state["prompt"] = "你好"
        iter = engine.generate(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)

    def test_chat_case(self):
        os.environ["LLM_TAG"] = "llm_llamacpp"
        engine: LLamacppLLM = LLMEnvInit.initLLMEngine()
        engine.args.llm_chat_system = "你是一个中国人,请用中文回答。回答限制在1-5句话内。要友好、乐于助人且简明扼要。保持对话简短而甜蜜。只用纯文本回答，不要包含链接或其他附加内容。不要回复计算机代码以及数学公式。"
        self.session.ctx.state["prompt"] = "你好"
        iter = engine.chat_completion(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)

        engine.args.llm_chat_system = ""
        self.session.ctx.state["prompt"] = "你好"
        iter = engine.chat_completion(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)

    def test_chat_vision_img(self):
        self.llm: LLamacppLLM = EngineFactory.get_engine_by_tag(
            EngineClass, self.llm_tag, **self.kwargs
        )
        self.llm.args.llm_stream = bool(self.stream)
        self.llm.args.llm_chat_system = ""
        self.session.ctx.state["prompt"] = [
            {"type": "text", "text": self.prompt},
            {
                "type": "image_url",
                "image_url": {
                    # "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
                    "url": "https://www.barnorama.com/wp-content/uploads/2016/12/03-Confusing-Pictures.jpg",
                },
            },
        ]
        iter = self.llm.chat_completion(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)

    def test_chat_vision_local_img(self):
        img_file = os.path.join(TEST_DIR, "img_files", "03-Confusing-Pictures.jpg")
        img_file = os.getenv("IMG_FILE", img_file)

        self.llm: LLamacppLLM = EngineFactory.get_engine_by_tag(
            EngineClass, self.llm_tag, **self.kwargs
        )
        self.llm.args.llm_stream = bool(self.stream)
        self.llm.args.llm_chat_system = ""

        data_uri = image_to_base64_data_uri(img_file)
        self.session.ctx.state["prompt"] = [
            {"type": "text", "text": self.prompt},
            {"type": "image_url", "image_url": {"url": data_uri}},
        ]
        iter = self.llm.chat_completion(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)

    def test_chat_vision_local_img_env_init(self):
        img_file = os.path.join(TEST_DIR, "img_files", "03-Confusing-Pictures.jpg")
        img_file = os.getenv("IMG_FILE", img_file)

        engine: LLamacppLLM = LLMEnvInit.initLLMEngine()
        engine.args.llm_stream = bool(self.stream)
        engine.args.llm_chat_system = ""
        # engine.args.llm_chat_system = "你是一个中国人,请用中文回答。回答限制在1-5句话内。要友好、乐于助人且简明扼要。保持对话简短而甜蜜。只用纯文本回答，不要包含链接或其他附加内容。不要回复计算机代码以及数学公式。"

        data_uri = image_to_base64_data_uri(img_file)
        self.session.ctx.state["prompt"] = [
            {"type": "text", "text": self.prompt},
            {"type": "image_url", "image_url": {"url": data_uri}},
        ]
        iter = engine.chat_completion(self.session)
        for item in iter:
            print(item)
            self.assertGreater(len(item), 0)



import os
import logging
import asyncio
import sys
import uuid

from apipeline.frames.control_frames import EndFrame
from apipeline.pipeline.task import PipelineTask

from src.processors.voice.base import VoiceProcessorBase
from src.processors.image.base import ImageGenProcessor
from src.processors.image import get_image_gen_processor
from src.modules.vision.ocr import VisionOCREnvInit
from src.modules.vision.detector import VisionDetectorEnvInit
from src.processors.ai_processor import AIProcessor
from src.processors.vision.vision_processor import MockVisionProcessor
from src.processors.speech.asr.base import ASRProcessorBase
from src.processors.llm.base import LLMProcessor
from src.processors.speech.tts.base import TTSProcessorBase
from src.modules.speech.vad_analyzer import VADAnalyzerEnvInit
from src.modules.speech.asr import ASREnvInit
from src.core.llm import LLMEnvInit
from src.modules.speech.tts import TTSEnvInit
from src.types.ai_conf import (
    TOGETHER_LLM_MODEL,
    TOGETHER_LLM_URL,
    ASRConfig,
    LLMConfig,
    TTSConfig,
    AIConfig,
)
from src.common import interface
from src.common.factory import EngineClass
from src.common.types import BotRunArgs
from src.common.interface import IBot, IVisionDetector
from src.common.session import Session
from src.common.types import SessionCtx

from dotenv import load_dotenv

load_dotenv(override=True)


class AIBot(IBot):
    r"""
    use ai bot config
    !TODONE: need config processor with bot config (redefine api params) @weedge
    bot config: Dict[str, Dict[str,Any]]
    e.g. {"llm":{"key":val,"tag":TAG,"args":{}}, "tts":{"key":val,"tag":TAG,"args":{}}}
    !TIPS: RTVI config options can transfer to ai bot config

    !NOTE:
    use multiprocessing pipe to run bot with unix socket, bot __init__ to new a bot obj must be serializable (pickle); or wraper a func, don't use bot obj methed.
    """

    def __init__(self, **args) -> None:
        self.args = BotRunArgs(**args)
        if self.args.bot_name is None or len(self.args.bot_name) == 0:
            self.args.bot_name = self.__class__.__name__

        self.task: PipelineTask | None = None
        self.session = Session(**SessionCtx(uuid.uuid4()).__dict__)

        self._bot_config_list = self.args.bot_config_list
        self._bot_config = self.args.bot_config
        self._handle_sigint = self.args.handle_sigint

    def init_bot_config(self):
        try:
            logging.debug(f"args.bot_config: {self.args.bot_config}")
            self._bot_config: AIConfig = AIConfig(**self.args.bot_config)
            if self._bot_config.llm is None:
                self._bot_config.llm = LLMConfig()
        except Exception as e:
            raise Exception(f"Failed to parse bot configuration: {e}")
        logging.info(f"ai bot_config: {self._bot_config}")

    def bot_config(self):
        return self._bot_config

    def run(self):
        asyncio.run(self.try_run())

    async def try_run(self):
        try:
            await self.arun()
        except asyncio.CancelledError:
            logging.info("CancelledError, Exiting!")
            if self.task is not None:
                await self.task.queue_frame(EndFrame())
        except KeyboardInterrupt:
            logging.warning("Ctrl-C detected. Exiting!")
            if self.task is not None:
                await self.task.queue_frame(EndFrame())
        except Exception as e:
            logging.error(f"run error: {e}", exc_info=True)

    async def arun(self):
        pass

    def get_vad_analyzer(self) -> interface.IVADAnalyzer | EngineClass:
        vad_analyzer: interface.IVADAnalyzer | EngineClass = None
        if (
            self._bot_config.vad
            and self._bot_config.vad.tag
            and len(self._bot_config.vad.tag) > 0
            and self._bot_config.vad.args
        ):
            vad_analyzer = VADAnalyzerEnvInit.getEngine(
                self._bot_config.vad.tag, **self._bot_config.vad.args
            )
        else:
            vad_analyzer = VADAnalyzerEnvInit.initVADAnalyzerEngine()
        return vad_analyzer

    def get_asr_processor(self) -> ASRProcessorBase:
        asr_processor: ASRProcessorBase | None = None
        if (
            self._bot_config.asr
            and self._bot_config.asr.tag
            and self._bot_config.asr.tag == "deepgram_asr_processor"
            and self._bot_config.asr.args
        ):
            from src.processors.speech.asr.deepgram_asr_processor import DeepgramAsrProcessor

            asr_processor = DeepgramAsrProcessor(
                api_key=os.getenv("DEEPGRAM_API_KEY"), **self._bot_config.asr.args
            )
        else:
            # use asr engine processor
            from src.processors.speech.asr.asr_processor import ASRProcessor

            asr: interface.IAsr | EngineClass | None = None
            if self._bot_config.asr and self._bot_config.asr.tag and self._bot_config.asr.args:
                asr = ASREnvInit.getEngine(self._bot_config.asr.tag, **self._bot_config.asr.args)
            else:
                logging.info("use default asr engine processor")
                asr = ASREnvInit.initASREngine()
                self._bot_config.asr = ASRConfig(tag=asr.SELECTED_TAG, args=asr.get_args_dict())
            asr_processor = ASRProcessor(asr=asr, session=self.session)
        return asr_processor

    def get_vision_llm_processor(self, llm_config: LLMConfig | None = None) -> LLMProcessor:
        """
        get local vision llm
        """
        from src.processors.vision.vision_processor import VisionProcessor

        if not llm_config:
            llm_config = self._bot_config.vision_llm
        if "mock" in llm_config.tag:
            llm_processor = MockVisionProcessor()
        else:
            logging.debug(f"init engine llm processor tag: {llm_config.tag}")
            llm_engine = LLMEnvInit.initLLMEngine(llm_config.tag, llm_config.args)
            llm_processor = VisionProcessor(llm_engine, self.session)
        return llm_processor

    def get_remote_llm_processor(self, llm: LLMConfig | None = None) -> LLMProcessor:
        """
        get remote llm
        """
        if not llm:
            llm = self._bot_config.llm
        if llm and llm.tag and "google" in llm.tag:
            llm_processor = self.get_google_llm_processor(llm)
        elif llm and llm.tag and "litellm" in llm.tag:
            llm_processor = self.get_litellm_processor(llm)
        elif llm and llm.tag and "openai" in llm.tag:
            llm_processor = self.get_openai_llm_processor(llm)
        else:
            llm_processor = self.get_google_llm_processor(llm)
        return llm_processor

    def get_openai_llm_processor(self, llm: LLMConfig | None = None) -> LLMProcessor:
        from src.processors.llm.openai_llm_processor import (
            OpenAILLMProcessor,
            OpenAIGroqLLMProcessor,
        )

        if not llm:
            llm = self._bot_config.llm
        # default use openai llm processor
        api_key = os.environ.get("OPENAI_API_KEY")
        if llm:
            if "groq" in llm.base_url:
                # https://console.groq.com/docs/models
                api_key = os.environ.get("GROQ_API_KEY")
                llm_processor = OpenAIGroqLLMProcessor(
                    model=llm.model,
                    base_url=llm.base_url,
                    api_key=api_key,
                )
                return llm_processor
            elif "together" in llm.base_url:
                # https://docs.together.ai/docs/chat-models
                api_key = os.environ.get("TOGETHER_API_KEY")
            else:
                llm.base_url = TOGETHER_LLM_URL
                llm.model = TOGETHER_LLM_MODEL
                api_key = os.environ.get("TOGETHER_API_KEY")

        llm_processor = OpenAILLMProcessor(
            model=llm.model,
            base_url=llm.base_url,
            api_key=api_key,
        )
        return llm_processor

    def get_google_llm_processor(self, llm: LLMConfig) -> LLMProcessor:
        from src.processors.llm.google_llm_processor import GoogleAILLMProcessor

        llm_config = llm
        if llm_config and llm_config.args:
            llm_processor = GoogleAILLMProcessor(**llm_config.args)
        else:
            logging.info("use default google llm processor")
            api_key = os.environ.get("GOOGLE_API_KEY")
            model = os.environ.get("GOOGLE_LLM_MODEL", "gemini-1.5-flash-latest")
            llm_processor = GoogleAILLMProcessor(
                api_key=api_key,
                model=model,
            )

        return llm_processor

    def get_litellm_processor(self, llm: LLMConfig) -> LLMProcessor:
        from src.processors.llm.litellm_processor import LiteLLMProcessor

        llm_processor = LiteLLMProcessor(model=llm.model, set_verbose=False)
        return llm_processor

    def get_llm_processor(self, llm: LLMConfig | None = None) -> LLMProcessor:
        if not llm:
            llm = self._bot_config.llm
        if llm and llm.tag and "vision" in llm.tag:
            # engine llm processor(just support vision model, other TODO):
            # (llm_llamacpp, llm_personalai_proxy, llm_transformers etc..)
            llm_processor = self.get_vision_llm_processor(llm)
        else:
            llm_processor = self.get_remote_llm_processor(llm)
        return llm_processor

    def get_vision_annotate_processor(self) -> AIProcessor:
        from src.processors.vision.annotate_processor import AnnotateProcessor

        detector: IVisionDetector | EngineClass = VisionDetectorEnvInit.initVisionDetectorEngine(
            self._bot_config.vision_detector.tag, self._bot_config.vision_detector.args
        )
        processor = AnnotateProcessor(detector, self.session)
        return processor

    def get_vision_detect_processor(self) -> AIProcessor:
        from src.processors.vision.detect_processor import DetectProcessor

        detector = VisionDetectorEnvInit.initVisionDetectorEngine(
            self._bot_config.vision_detector.tag, self._bot_config.vision_detector.args
        )

        desc, out_desc = "", ""
        if "desc" in self._bot_config.vision_detector.args:
            desc = self._bot_config.vision_detector.args["desc"]
        if "out_desc" in self._bot_config.vision_detector.args:
            out_desc = self._bot_config.vision_detector.args["out_desc"]
        if desc and out_desc:
            processor = DetectProcessor(
                detected_text=desc,
                out_detected_text=out_desc,
                detector=detector,
                session=self.session,
            )
        elif desc:
            processor = DetectProcessor(detected_text=desc, detector=detector, session=self.session)
        elif out_desc:
            processor = DetectProcessor(
                out_detected_text=out_desc, detector=detector, session=self.session
            )
        else:
            processor = DetectProcessor(detector=detector, session=self.session)

        return processor

    def get_vision_ocr_processor(self) -> AIProcessor:
        from src.processors.vision.ocr_processor import OCRProcessor

        ocr = VisionOCREnvInit.initVisionOCREngine(
            self._bot_config.vision_ocr.tag, self._bot_config.vision_ocr.args
        )
        processor = OCRProcessor(ocr=ocr, session=self.session)
        return processor

    def get_moshi_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        if not llm:
            llm = self._bot_config.voice_llm
        if llm and llm.tag and "moshi" in llm.tag:
            from src.processors.voice.moshi_voice_processor import (
                MoshiVoiceOpusStreamProcessor,
                MoshiVoiceProcessor,
            )

            if "moshi_opus" in llm.tag:
                if llm.args:
                    llm_processor = MoshiVoiceOpusStreamProcessor(**llm.args)
                else:
                    llm_processor = MoshiVoiceOpusStreamProcessor()
            else:
                if llm.args:
                    llm_processor = MoshiVoiceProcessor(**llm.args)
                else:
                    llm_processor = MoshiVoiceProcessor()
        else:
            from src.processors.voice.moshi_voice_processor import VoiceOpusStreamEchoProcessor

            llm_processor = VoiceOpusStreamEchoProcessor()
        return llm_processor

    def get_text_glm_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from src.processors.voice.glm_voice_processor import GLMTextVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = GLMTextVoiceProcessor(**llm.args)
        else:
            llm_processor = GLMTextVoiceProcessor()
        return llm_processor

    def get_audio_glm_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        from src.processors.voice.glm_voice_processor import GLMAudioVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        if llm.args:
            llm_processor = GLMAudioVoiceProcessor(**llm.args)
        else:
            llm_processor = GLMAudioVoiceProcessor()
        return llm_processor

    def get_audio_freeze_omni_voice_processor(
        self, llm: LLMConfig | None = None
    ) -> VoiceProcessorBase:
        cur_dir = os.path.dirname(__file__)
        if bool(os.getenv("ACHATBOT_PKG", "")):
            sys.path.insert(1, os.path.join(cur_dir, "../../FreezeOmni"))
        else:
            sys.path.insert(1, os.path.join(cur_dir, "../../../deps/FreezeOmni"))

        from src.processors.voice.freeze_omni_voice_processor import FreezeOmniVoiceProcessor

        if not llm:
            llm = self._bot_config.voice_llm
        # TODO: need create inference_pipeline_pool and tts pool
        if llm.args:
            llm_processor = FreezeOmniVoiceProcessor(**llm.args)
        else:
            llm_processor = FreezeOmniVoiceProcessor()
        return llm_processor

    def get_openai_voice_processor(self, llm: LLMConfig | None = None) -> VoiceProcessorBase:
        # TODO: use openai reatime voice api
        pass

    def get_tts_processor(self) -> TTSProcessorBase:
        tts_processor: TTSProcessorBase | None = None
        if self._bot_config.tts and self._bot_config.tts.tag and self._bot_config.tts.args:
            if self._bot_config.tts.tag == "elevenlabs_tts_processor":
                from src.processors.speech.tts.elevenlabs_tts_processor import (
                    ElevenLabsTTSProcessor,
                )

                tts_processor = ElevenLabsTTSProcessor(**self._bot_config.tts.args)
            elif self._bot_config.tts.tag == "cartesia_tts_processor":
                from src.processors.speech.tts.cartesia_tts_processor import CartesiaTTSProcessor

                tts_processor = CartesiaTTSProcessor(**self._bot_config.tts.args)
            else:
                # use tts engine processor
                from src.processors.speech.tts.tts_processor import TTSProcessor

                tts = TTSEnvInit.getEngine(self._bot_config.tts.tag, **self._bot_config.tts.args)
                self._bot_config.tts.tag = (tts.SELECTED_TAG,)
                self._bot_config.tts.args = tts.get_args_dict()
                tts_processor = TTSProcessor(tts=tts, session=self.session)
        else:
            # default tts engine processor
            from src.processors.speech.tts.tts_processor import TTSProcessor

            logging.info("use default tts engine processor")
            tag = None
            if self._bot_config.tts and self._bot_config.tts.tag:
                tag = self._bot_config.tts.tag
            tts = TTSEnvInit.initTTSEngine(tag)
            self._bot_config.tts = TTSConfig(tag=tts.SELECTED_TAG, args=tts.get_args_dict())
            tts_processor = TTSProcessor(tts=tts, session=self.session)

        return tts_processor

    def get_image_gen_processor(self) -> ImageGenProcessor:
        if not self._bot_config.img_gen or not self._bot_config.img_gen.args:
            raise Exception("need img_gen args params")
        return get_image_gen_processor(
            self._bot_config.img_gen.tag, **self._bot_config.img_gen.args
        )


class AIRoomBot(AIBot):
    def __init__(self, **args) -> None:
        super().__init__(**args)


class AIChannelBot(AIBot):
    def __init__(self, **args) -> None:
        super().__init__(**args)



import json
import torch
import threading
import queue
import subprocess
import concurrent.futures

import numpy as np
import torchaudio.compliance.kaldi as k

from silero_vad.model import load_silero_vad
from silero_vad.utils_vad import VADIterator

class VAD:
    """
    首先使用声学 VAD 模块来检测流式演讲的起点。
    当VAD被触发时，语音流将被逐块发送到Freeze-Omni，并在LLM最后一层之后添加一个额外的分类层来预测不同的状态。
    这里定义了三种状态:
    - 状态0表示当前LLM可以继续接收语音，
    - 状态1或2表示当前块是语音结束。
        - 状态1表示用户将中断对话，LLM将执行新的生成阶段，
        - 状态2表示无需中断对话。
        这两种状态都将停止向 Freeze-Omni 发送语音流并重置 VAD 模块。
    """
    def __init__(self, cache_history=10, last_chunk_size=6):
        self.chunk_size = 16
        self.chunk_overlap = 3
        self.feat_dim = 80
        self.frame_size = 400
        self.frame_shift = 160
        self.frame_overlap = self.frame_size - self.frame_shift
        self.CHUNK = self.frame_shift * self.chunk_size
        assert cache_history >= last_chunk_size, "cache_history must >= last_chunk_size"
        self.cache_history = cache_history
        self.last_chunk_size = last_chunk_size
        self.in_dialog = False

        with torch.no_grad():
            self.load_vad()
            self.reset_vad()
    
    def get_chunk_size(self):
        return self.CHUNK

    def load_vad(self):
        self.vad_model = load_silero_vad()
        self.vad_model.eval()
        # generate vad itertator
        self.vad_iterator = VADIterator(self.vad_model, 
                                        threshold=0.8, 
                                        sampling_rate=16000, 
                                        min_silence_duration_ms=2000, 
                                        speech_pad_ms=30)
        self.vad_iterator.reset_states()

    def reset_vad(self):
        # reset all parms
        self.input_chunk = torch.zeros([1, self.chunk_size + self.chunk_overlap, self.feat_dim])
        self.input_sample = torch.zeros([1, self.CHUNK + self.frame_overlap , 1])
        # chunck feat history cache also use ring buffer to do :)
        self.history = torch.zeros([self.cache_history, self.chunk_size + self.chunk_overlap, self.feat_dim])
        self.vad_iterator.reset_states()
        self.in_dialog = False
    
    def run_vad_iterator(self, audio):
        speech_dict_out = None
        # split into chunk with 512
        for i in range(len(audio) // 512):
            speech_dict = self.vad_iterator(audio[i * 512: (i + 1) * 512], return_seconds=True)
            if speech_dict is not None:
                speech_dict_out = speech_dict
        return speech_dict_out
    
    def predict(self,
                audio: np.ndarray):
        """
        Predict the Voice Activity Detection (VAD) status and return related features.

        Parameters:
        - audio (torch.Tensor): A 1D or 2D tensor representing the input audio signal.

        Returns:
        - return_dict (dict): A dictionary containing the VAD status and related features.
            - 'status' (str): The current VAD status, which can be 'sl' (speech start), 
                              'cl' (speech continue), or 'el' (speech end).
            - 'feature_last_chunk' (list of list of float): The feature of the last chunks.
            - 'feature' (list of list of float): The feature of the current chunk of audio.
            - 'history_feature' (list of list of list of float): The cached features of previous chunks.
        
        """

        # 1. Converts the input audio tensor to the appropriate format.
        # 2. Computes the filter bank features (fbank) for the audio.
        # 3. Updates the input chunk and history based on the new audio segment.
        # 4. Determines the VAD status by running the VAD iterator on the audio.
        # 5. Populates the return dictionary with the VAD status and related features.

        return_dict = {}
        return_dict['status'] = None
        with torch.no_grad():
            # get fbank feature
            audio = torch.tensor(audio)
            sample_data = audio.reshape(1, -1, 1)[:, :, :1] * 32768
            self.input_sample[:, :self.frame_overlap , :] = self.input_sample[:, -self.frame_overlap:, :].clone()
            self.input_sample[:, self.frame_overlap:, :] = sample_data
            # compute kaldi style feature
            xs = k.fbank(waveform = self.input_sample.squeeze(-1), dither=0, 
                        frame_length=25, frame_shift=10, num_mel_bins=self.feat_dim)
            self.input_chunk[:, :self.chunk_overlap, :] = self.input_chunk[:, -self.chunk_overlap:, :].clone()
            self.input_chunk[:, self.chunk_overlap:, :] = xs.squeeze(0)

            # get vad status
            if self.in_dialog:
                speech_dict = self.run_vad_iterator(audio.reshape(-1))
                if speech_dict is not None and "end" in speech_dict:
                    return_dict['status'] = 'el'
                    # reset state
                    self.vad_iterator.reset_states()
                else:
                    return_dict['status'] = 'cl'
            if not self.in_dialog:
                speech_dict = self.run_vad_iterator(audio.reshape(-1))
                if speech_dict is not None and "start" in speech_dict:
                    return_dict['status'] = 'sl'
                    self.in_dialog = True
                    # self.vad_iterator.reset_states()
                else:  
                    # cache fbank feature
                    # << 1
                    self.history[:-1] = self.history[1:].clone()
                    # last history = input chunk
                    self.history[-1:] = self.input_chunk

            # return dict
            if return_dict['status'] == 'sl':
                # copy last chunk size chunks
                return_dict['feature_last_chunk'] = self.history[-self.last_chunk_size:].unsqueeze(1)
                return_dict['feature'] = self.input_chunk
                return_dict['history_feature'] = self.history
            elif return_dict['status'] == 'cl' or return_dict['status'] == 'el':
                return_dict['feature_last_chunk'] = None
                return_dict['feature'] = self.input_chunk
                return_dict['history_feature'] = self.history

        return return_dict



import asyncio
import os
import logging

import unittest
from livekit import rtc
from apipeline.pipeline.pipeline import Pipeline
from apipeline.processors.logger import FrameLogger
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.frames.control_frames import EndFrame
from apipeline.frames.data_frames import TextFrame, DataFrame, AudioRawFrame, ImageRawFrame
from apipeline.processors.output_processor import OutputFrameProcessor

from src.modules.speech.asr import ASREnvInit
from src.processors.speech.asr.asr_processor import ASRProcessor
from src.services.help.livekit_room import LivekitRoom
from src.services.help import RoomManagerEnvInit
from src.common.types import LivekitParams
from src.common.logger import Logger
from src.transports.livekit import LivekitTransport
from src.types.frames.data_frames import TranscriptionFrame

from dotenv import load_dotenv

load_dotenv(override=True)

r"""
LIVEKIT_BOT_NAME=chat-bot \
    python -m unittest test.integration.processors.test_livekit_asr_processor.TestASRProcessor

LIVEKIT_BOT_NAME=chat-bot  \
    ASR_LANG=zn \
    ASR_MODEL_NAME_OR_PATH=./models/FunAudioLLM/SenseVoiceSmall \
    ASR_VERBOSE=True \
    ASR_TAG=sense_voice_asr
    python -m unittest test.integration.processors.test_livekit_asr_processor.TestASRProcessor
"""


class TestASRProcessor(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False)
        cls.room_name = os.getenv("ROOM_NAME", "chat-room")

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        self.texts = []

        self.livekit_room: LivekitRoom = RoomManagerEnvInit.initEngine("livekit_room")
        self.asr = ASREnvInit.initASREngine()

    async def asyncTearDown(self):
        await self.livekit_room.close_session()

    async def test_tts_daily_output(self):
        token = await self.livekit_room.gen_token(self.room_name)
        transport = LivekitTransport(
            token,
            params=LivekitParams(
                audio_in_enabled=True,
                audio_in_sample_rate=48000,
            ),
        )
        self.assertGreater(len(transport.event_names), 0)

        asr_processor = ASRProcessor(asr=self.asr)
        out_processor = OutputFrameProcessor(cb=self.sink_callback)

        self.task = PipelineTask(
            Pipeline(
                [
                    transport.input_processor(),
                    # FrameLogger(include_frame_types=[AudioRawFrame]),
                    asr_processor,
                    FrameLogger(include_frame_types=[TranscriptionFrame]),
                    out_processor,
                ]
            ),
            params=PipelineParams(allow_interruptions=True),
        )

        transport.add_event_handler("on_first_participant_joined", self.on_first_participant_joined)
        transport.add_event_handler("on_participant_disconnected", self.on_participant_disconnected)
        transport.add_event_handler("on_disconnected", self.on_disconnected)
        # sometime don't get on_connection_state_changed when disconnected
        transport.add_event_handler("on_connection_state_changed", self.on_connection_state_changed)

        runner = PipelineRunner()
        try:
            await asyncio.wait_for(runner.run(self.task), int(os.getenv("RUN_TIMEOUT", "60")))
        except asyncio.TimeoutError:
            logging.warning("Test run timeout. Exiting")
            await self.task.queue_frame(EndFrame())

    async def sink_callback(self, frame: DataFrame):
        print(f"sink_callback ----> print frame: {frame}")
        if not isinstance(frame, TextFrame):
            return

        self.texts.append(frame.text)
        if len(self.texts) == 2:  # asr 2 times (speech to text)
            print("sink_callback ----> send end frame")
            await self.task.queue_frame(EndFrame())
            return

    async def on_first_participant_joined(
        self, transport: LivekitTransport, participant: rtc.RemoteParticipant
    ):
        logging.debug(f"transport---->{transport}")
        logging.debug(f"participant---->{participant}")
        participant_name = participant.name if participant.name else participant.identity
        await transport.send_message(
            f"hello,你好，{participant_name}, 我是机器人。",
            participant_id=participant.identity,
        )

    async def on_participant_disconnected(
        self, transport: LivekitTransport, participant: rtc.RemoteParticipant
    ):
        logging.info(f"Partcipant {participant} left. Exiting.")

    async def on_disconnected(self, transport: LivekitTransport, reason: str):
        logging.info("disconnected reason %s, Exiting." % reason)
        await self.task.queue_frame(EndFrame())

    async def on_connection_state_changed(
        self, transport: LivekitTransport, state: rtc.ConnectionState
    ):
        logging.info("connection state %s " % state)
        if state == rtc.ConnectionState.CONN_DISCONNECTED:
            await self.task.queue_frame(EndFrame())



from io import BytesIO
import logging
from threading import Thread

from PIL import Image
import torch

try:
    from qwen_vl_utils import process_vision_info
    from transformers import (
        AutoModelForCausalLM,
        AutoProcessor,
        TextIteratorStreamer,
        GenerationConfig,
    )
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Molmo multimodal language models., you need to `pip install achatbot[llm_transformers_manual_vision_molmo]`,"
        "use awq model need to `pip install achatbot[llm_transformers_manual_vision_molmo,autoawq]`"
    )
    raise Exception(f"Missing module: {e}")


from achatbot.common.session import Session
from achatbot.types.speech.language import TO_LLM_LANGUAGE
from achatbot.types.llm.transformers import TransformersLMArgs
from .base import TransformersBaseLLM


class TransformersManualVisionMolmoLLM(TransformersBaseLLM):
    r"""
    no chat template, no chat history, just image with text prompt
    """

    TAG = "llm_transformers_manual_vision_molmo"

    def __init__(self, **args) -> None:
        self.args = TransformersLMArgs(**args)
        if self.args.lm_torch_dtype != "auto":
            self.torch_dtype = getattr(torch, self.args.lm_torch_dtype)
        else:
            self.torch_dtype = "auto"

        if self.args.lm_device_map:
            self._model = AutoModelForCausalLM.from_pretrained(
                self.args.lm_model_name_or_path,
                torch_dtype=self.args.lm_torch_dtype,
                #!NOTE: https://github.com/huggingface/transformers/issues/20896
                # device_map for multi cpu/gpu with accelerate
                device_map=self.args.lm_device_map,
                attn_implementation=self.args.lm_attn_impl,
                trust_remote_code=True,
            )
        else:
            self._model = (
                AutoModelForCausalLM.from_pretrained(
                    self.args.lm_model_name_or_path,
                    torch_dtype=self.args.lm_torch_dtype,
                    attn_implementation=self.args.lm_attn_impl,
                    trust_remote_code=True,
                )
                .eval()
                .to(self.args.lm_device)
            )

        # The default range for the number of visual tokens per image in the model is 4-16384.
        # You can set min_pixels and max_pixels according to your needs, such as a
        # token count range of 256-1280, to balance speed and memory usage.
        self._processor = AutoProcessor.from_pretrained(
            self.args.lm_model_name_or_path,
            min_pixels=256 * 28 * 28,
            max_pixels=1280 * 28 * 28,
            trust_remote_code=True,
        )

        self.warmup()

    def warmup(self):
        dummy_input_text = self.args.warnup_prompt

        dummy_image = Image.new(mode="RGB", size=(300, 300))
        img_obj = None
        with BytesIO() as buffered:
            dummy_image.save(buffered, "JPEG")
            img_obj = Image.open(BytesIO(buffered.getvalue()))

        # Preparation for inference
        inputs = self._processor.process(
            images=[img_obj],
            text=dummy_input_text,
        )
        # move inputs to the correct device and make a batch of size 1
        model_inputs = {k: v.to(self._model.device).unsqueeze(0) for k, v in inputs.items()}

        streamer = TextIteratorStreamer(
            self._processor.tokenizer, skip_prompt=True, skip_special_tokens=True
        )

        warmup_gen_kwargs = dict(
            streamer=streamer,
            tokenizer=self._processor.tokenizer,
        )
        warmup_gen_args = (
            model_inputs,
            GenerationConfig(
                do_sample=self.args.lm_gen_do_sample,
                top_k=self.args.lm_gen_top_k,
                top_p=self.args.lm_gen_top_p,
                temperature=self.args.lm_gen_temperature,
                repetition_penalty=self.args.lm_gen_repetition_penalty,
                min_new_tokens=self.args.lm_gen_min_new_tokens,
                max_new_tokens=self.args.lm_gen_max_new_tokens,
                stop_strings="<|endoftext|>",
            ),
        )
        self._warmup(
            target=self._model.generate_from_batch,
            args=warmup_gen_args,
            kwargs=warmup_gen_kwargs,
            streamer=streamer,
        )

    def generate(self, session: Session):
        prompt = session.ctx.state["prompt"]
        text = ""
        image_inputs = None

        if isinstance(prompt, str):
            text = prompt
        elif isinstance(prompt, tuple):
            prompt, language_code = prompt
            if isinstance(prompt, str):
                if TO_LLM_LANGUAGE[language_code] == "zh":
                    prompt = "请用中文回复。" + prompt
                else:
                    prompt = (
                        f"Please reply to my message in {TO_LLM_LANGUAGE[language_code]}. " + prompt
                    )
            text = prompt
        elif isinstance(prompt, list):
            # no chat template so get image with textpromt
            image_inputs, _ = process_vision_info(
                [{"role": self.args.user_role, "content": prompt}]
            )
            for item in prompt:
                if "type" in item and item["type"] == "text":
                    text += item["text"]
        else:
            raise Exception(f"Unsupported prompt type: {type(prompt)}")

        # Preparation for inference
        inputs = self._processor.process(
            images=image_inputs,
            text=text,
        )
        # move inputs to the correct device and make a batch of size 1
        model_inputs = {k: v.to(self._model.device).unsqueeze(0) for k, v in inputs.items()}

        streamer = TextIteratorStreamer(
            self._processor.tokenizer, skip_prompt=True, skip_special_tokens=True
        )

        gen_kwargs = dict(
            streamer=streamer,
            tokenizer=self._processor.tokenizer,
        )
        gen_args = (
            model_inputs,
            GenerationConfig(
                do_sample=self.args.lm_gen_do_sample,
                top_k=self.args.lm_gen_top_k,
                top_p=self.args.lm_gen_top_p,
                temperature=self.args.lm_gen_temperature,
                repetition_penalty=self.args.lm_gen_repetition_penalty,
                min_new_tokens=self.args.lm_gen_min_new_tokens,
                max_new_tokens=self.args.lm_gen_max_new_tokens,
                stop_strings="<|endoftext|>",
            ),
        )
        thread = Thread(target=self._model.generate_from_batch, args=gen_args, kwargs=gen_kwargs)
        thread.start()

        generated_text = ""
        for new_text in streamer:
            generated_text += new_text
            yield new_text



import glob
import os
import sys
import logging
import json

import numpy as np
import torch
import sounddevice as sd
import typer

from src.common.types import MODELS_DIR

app = typer.Typer()

"""
# if not espeak-ng, brew update
brew install espeak-ng
huggingface-cli download hexgrad/Kokoro-82M --quiet --local-dir ./models/Kokoro82M
touch models/__init__.py
python -m demo.tts_kokoro run_torch_kokoro
"""
try:
    sys.path.insert(1, os.path.join(MODELS_DIR, "Kokoro82M"))
    from models.Kokoro82M.kokoro import generate
    from models.Kokoro82M.models import build_model
except ModuleNotFoundError as e:
    logging.error("In order to use kokoro-tts, you need to `pip install achatbot[tts_kokoro]`.")
    raise Exception(f"Missing module: {e}")

"""
brew install espeak-ng
python -m demo.tts_kokoro export_pytorch_voices_to_json
python -m demo.tts_kokoro run_onnx_kokoro
"""
try:
    from kokoro_onnx import Kokoro, EspeakConfig
except ModuleNotFoundError as e:
    logging.error(
        "In order to use kokoro-tts with onnx, you need to `pip install achatbot[tts_onnx_kokoro]`."
    )
    raise Exception(f"Missing module: {e}")


@app.command("run_torch_kokoro")
def run_torch_kokoro(
    text="How could I know? It's an unanswerable question. Like asking an unborn child if they'll lead a good life. They haven't even been born.",
):
    device = "cuda" if torch.cuda.is_available() else "cpu"
    MODEL = build_model(os.path.join(MODELS_DIR, "Kokoro82M/kokoro-v0_19.pth"), device)

    total_params = 0
    for key, model in MODEL.items():
        logging.info(f"{key} Model: {model}")
        params = sum(p.numel() for p in model.parameters())
        total_params += params
        model_million_params = params / 1e6
        logging.info(f"{key} Model has {model_million_params:.3f} million parameters")

    model_million_params = total_params / 1e6
    logging.info(f"Model total has {model_million_params:.3f} million parameters")

    VOICE_NAME = [
        "af",  # Default voice is a 50-50 mix of Bella & Sarah
        "af_bella",
        "af_sarah",
        "am_adam",
        "am_michael",
        "bf_emma",
        "bf_isabella",
        "bm_george",
        "bm_lewis",
        "af_nicole",
        "af_sky",
    ][0]
    VOICEPACK = torch.load(
        os.path.join(MODELS_DIR, f"Kokoro82M/voices/{VOICE_NAME}.pt"), weights_only=True
    ).to(device)
    logging.info(f"Loaded voice: {VOICE_NAME}")

    audio_samples, out_ps = generate(MODEL, text, VOICEPACK, lang=VOICE_NAME[0])
    print(audio_samples.shape, audio_samples, out_ps)
    sd.play(audio_samples, 24000)
    sd.wait()


@app.command("export_pytorch_voices_to_json")
def export_pytorch_voices_to_json():
    """
    export pytorch voices to json for onnx
    """
    voices_json = {}
    voices_pt_file_list = glob.glob(os.path.join(MODELS_DIR, "Kokoro82M/voices/*.pt"))
    # logging.info(voices_pt_file_list)

    for voice_file in voices_pt_file_list:
        voice = os.path.splitext(os.path.basename(voice_file))[0]
        logging.info(f"voice {voice} file: {voice_file}")
        voice_data: np.ndarray = torch.load(voice_file).numpy()
        voices_json[voice] = voice_data.tolist()

    voice_json_file = os.path.join(MODELS_DIR, "Kokoro82M/kokoro-voices.json")
    with open(voice_json_file, "w") as f:
        json.dump(voices_json, f, indent=4)
        logging.info(f"dump to {voice_json_file}")


@app.command("run_onnx_kokoro")
def run_onnx_kokoro(
    text="How could I know? It's an unanswerable question. Like asking an unborn child if they'll lead a good life. They haven't even been born.",
):
    model_struct_stats_ckpt = os.path.join(MODELS_DIR, "Kokoro82M/kokoro-v0_19.onnx")
    voices_file = os.path.join(MODELS_DIR, "Kokoro82M/kokoro-voices.json")
    print(model_struct_stats_ckpt, voices_file)
    espeak_ng_lib_path = os.getenv("ESPEAK_NG_LIB_PATH", "/usr/local/lib/libespeak-ng.1.dylib")
    kokoro = Kokoro(
        model_struct_stats_ckpt,
        voices_file,
        espeak_config=EspeakConfig(lib_path=espeak_ng_lib_path),
    )
    audio_samples, sample_rate = kokoro.create(text, voice="af_sarah", speed=1.0, lang="en-us")
    print(audio_samples.shape, audio_samples, sample_rate)
    sd.play(audio_samples, sample_rate)
    sd.wait()


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s",
        handlers=[logging.StreamHandler()],
    )
    app()



from dataclasses import dataclass

import deps.ChatTTS.ChatTTS as ChatTTS


@dataclass
class ChatTTSArgs:
    source: str = "huggingface"  # "huggingface", "local", "custom"
    force_redownload: bool = False
    local_path: str = ""
    compile: bool = True
    device: str = None
    use_flash_attn: bool = False
    skip_refine_text: bool = False
    refine_text_only: bool = False
    params_refine_text: ChatTTS.Chat.RefineTextParams | None = None
    params_infer_code: ChatTTS.Chat.InferCodeParams | None = None
    use_decoder: bool = True
    do_text_normalization: bool = False
    lang: str = None
    tts_stream: bool = False



import os

os.environ["PYTOCH_ENABLE_MPS_FALLBACK"] = "1"  # for MPS device compatibility

import torch
import torch.nn.functional as F
import torchaudio

from f5_tts.infer.utils_infer import load_checkpoint, load_vocoder, save_spectrogram
from f5_tts.model import CFM, DiT, UNetT
from f5_tts.model.utils import convert_char_to_pinyin, get_tokenizer

device = "cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu"


# --------------------- Dataset Settings -------------------- #

target_sample_rate = 24000
n_mel_channels = 100
hop_length = 256
win_length = 1024
n_fft = 1024
mel_spec_type = "vocos"  # 'vocos' or 'bigvgan'
target_rms = 0.1

tokenizer = "pinyin"
dataset_name = "Emilia_ZH_EN"


# ---------------------- infer setting ---------------------- #

seed = None  # int | None

exp_name = "F5TTS_Base"  # F5TTS_Base | E2TTS_Base
ckpt_step = 1200000

nfe_step = 32  # 16, 32
cfg_strength = 2.0
ode_method = "euler"  # euler | midpoint
sway_sampling_coef = -1.0
speed = 1.0

if exp_name == "F5TTS_Base":
    model_cls = DiT
    model_cfg = dict(dim=1024, depth=22, heads=16, ff_mult=2, text_dim=512, conv_layers=4)

elif exp_name == "E2TTS_Base":
    model_cls = UNetT
    model_cfg = dict(dim=1024, depth=24, heads=16, ff_mult=4)

ckpt_path = f"ckpts/{exp_name}/model_{ckpt_step}.safetensors"
output_dir = "tests"

# [leverage https://github.com/MahmoudAshraf97/ctc-forced-aligner to get char level alignment]
# pip install git+https://github.com/MahmoudAshraf97/ctc-forced-aligner.git
# [write the origin_text into a file, e.g. tests/test_edit.txt]
# ctc-forced-aligner --audio_path "src/f5_tts/infer/examples/basic/basic_ref_en.wav" --text_path "tests/test_edit.txt" --language "zho" --romanize --split_size "char"
# [result will be saved at same path of audio file]
# [--language "zho" for Chinese, "eng" for English]
# [if local ckpt, set --alignment_model "../checkpoints/mms-300m-1130-forced-aligner"]

audio_to_edit = "src/f5_tts/infer/examples/basic/basic_ref_en.wav"
origin_text = "Some call me nature, others call me mother nature."
target_text = "Some call me optimist, others call me realist."
parts_to_edit = [
    [1.42, 2.44],
    [4.04, 4.9],
]  # stard_ends of "nature" & "mother nature", in seconds
fix_duration = [
    1.2,
    1,
]  # fix duration for "optimist" & "realist", in seconds

# audio_to_edit = "src/f5_tts/infer/examples/basic/basic_ref_zh.wav"
# origin_text = "对，这就是我，万人敬仰的太乙真人。"
# target_text = "对，那就是你，万人敬仰的太白金星。"
# parts_to_edit = [[0.84, 1.4], [1.92, 2.4], [4.26, 6.26], ]
# fix_duration = None  # use origin text duration


# -------------------------------------------------#

use_ema = True

if not os.path.exists(output_dir):
    os.makedirs(output_dir)

# Vocoder model
local = False
if mel_spec_type == "vocos":
    vocoder_local_path = "../checkpoints/charactr/vocos-mel-24khz"
elif mel_spec_type == "bigvgan":
    vocoder_local_path = "../checkpoints/bigvgan_v2_24khz_100band_256x"
vocoder = load_vocoder(vocoder_name=mel_spec_type, is_local=local, local_path=vocoder_local_path)

# Tokenizer
vocab_char_map, vocab_size = get_tokenizer(dataset_name, tokenizer)

# Model
model = CFM(
    transformer=model_cls(**model_cfg, text_num_embeds=vocab_size, mel_dim=n_mel_channels),
    mel_spec_kwargs=dict(
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        n_mel_channels=n_mel_channels,
        target_sample_rate=target_sample_rate,
        mel_spec_type=mel_spec_type,
    ),
    odeint_kwargs=dict(
        method=ode_method,
    ),
    vocab_char_map=vocab_char_map,
).to(device)

dtype = torch.float32 if mel_spec_type == "bigvgan" else None
model = load_checkpoint(model, ckpt_path, device, dtype=dtype, use_ema=use_ema)

# Audio
audio, sr = torchaudio.load(audio_to_edit)
if audio.shape[0] > 1:
    audio = torch.mean(audio, dim=0, keepdim=True)
rms = torch.sqrt(torch.mean(torch.square(audio)))
if rms < target_rms:
    audio = audio * target_rms / rms
if sr != target_sample_rate:
    resampler = torchaudio.transforms.Resample(sr, target_sample_rate)
    audio = resampler(audio)
offset = 0
audio_ = torch.zeros(1, 0)
edit_mask = torch.zeros(1, 0, dtype=torch.bool)
for part in parts_to_edit:
    start, end = part
    part_dur = end - start if fix_duration is None else fix_duration.pop(0)
    part_dur = part_dur * target_sample_rate
    start = start * target_sample_rate
    audio_ = torch.cat((audio_, audio[:, round(offset) : round(start)], torch.zeros(1, round(part_dur))), dim=-1)
    edit_mask = torch.cat(
        (
            edit_mask,
            torch.ones(1, round((start - offset) / hop_length), dtype=torch.bool),
            torch.zeros(1, round(part_dur / hop_length), dtype=torch.bool),
        ),
        dim=-1,
    )
    offset = end * target_sample_rate
# audio = torch.cat((audio_, audio[:, round(offset):]), dim = -1)
edit_mask = F.pad(edit_mask, (0, audio.shape[-1] // hop_length - edit_mask.shape[-1] + 1), value=True)
audio = audio.to(device)
edit_mask = edit_mask.to(device)

# Text
text_list = [target_text]
if tokenizer == "pinyin":
    final_text_list = convert_char_to_pinyin(text_list)
else:
    final_text_list = [text_list]
print(f"text  : {text_list}")
print(f"pinyin: {final_text_list}")

# Duration
ref_audio_len = 0
duration = audio.shape[-1] // hop_length

# Inference
with torch.inference_mode():
    generated, trajectory = model.sample(
        cond=audio,
        text=final_text_list,
        duration=duration,
        steps=nfe_step,
        cfg_strength=cfg_strength,
        sway_sampling_coef=sway_sampling_coef,
        seed=seed,
        edit_mask=edit_mask,
    )
    print(f"Generated mel: {generated.shape}")

    # Final result
    generated = generated.to(torch.float32)
    generated = generated[:, ref_audio_len:, :]
    gen_mel_spec = generated.permute(0, 2, 1)
    if mel_spec_type == "vocos":
        generated_wave = vocoder.decode(gen_mel_spec).cpu()
    elif mel_spec_type == "bigvgan":
        generated_wave = vocoder(gen_mel_spec).squeeze(0).cpu()

    if rms < target_rms:
        generated_wave = generated_wave * rms / target_rms

    save_spectrogram(gen_mel_spec[0].cpu().numpy(), f"{output_dir}/speech_edit_out.png")
    torchaudio.save(f"{output_dir}/speech_edit_out.wav", generated_wave, target_sample_rate)
    print(f"Generated wav: {generated_wave.shape}")



from typing import Dict, Optional
import torch
import torch.nn as nn
import torch.nn.functional as F
from matcha.hifigan.models import feature_loss, generator_loss, discriminator_loss
from cosyvoice.utils.losses import tpr_loss, mel_loss


class HiFiGan(nn.Module):
    def __init__(self, generator, discriminator, mel_spec_transform,
                 multi_mel_spectral_recon_loss_weight=45, feat_match_loss_weight=2.0,
                 tpr_loss_weight=1.0, tpr_loss_tau=0.04):
        super(HiFiGan, self).__init__()
        self.generator = generator
        self.discriminator = discriminator
        self.mel_spec_transform = mel_spec_transform
        self.multi_mel_spectral_recon_loss_weight = multi_mel_spectral_recon_loss_weight
        self.feat_match_loss_weight = feat_match_loss_weight
        self.tpr_loss_weight = tpr_loss_weight
        self.tpr_loss_tau = tpr_loss_tau

    def forward(
            self,
            batch: dict,
            device: torch.device,
    ) -> Dict[str, Optional[torch.Tensor]]:
        if batch['turn'] == 'generator':
            return self.forward_generator(batch, device)
        else:
            return self.forward_discriminator(batch, device)

    def forward_generator(self, batch, device):
        real_speech = batch['speech'].to(device)
        pitch_feat = batch['pitch_feat'].to(device)
        # 1. calculate generator outputs
        generated_speech, generated_f0 = self.generator(batch, device)
        # 2. calculate discriminator outputs
        y_d_rs, y_d_gs, fmap_rs, fmap_gs = self.discriminator(real_speech, generated_speech)
        # 3. calculate generator losses, feature loss, mel loss, tpr losses [Optional]
        loss_gen, _ = generator_loss(y_d_gs)
        loss_fm = feature_loss(fmap_rs, fmap_gs)
        loss_mel = mel_loss(real_speech, generated_speech, self.mel_spec_transform)
        if self.tpr_loss_weight != 0:
            loss_tpr = tpr_loss(y_d_rs, y_d_gs, self.tpr_loss_tau)
        else:
            loss_tpr = torch.zeros(1).to(device)
        loss_f0 = F.l1_loss(generated_f0, pitch_feat)
        loss = loss_gen + self.feat_match_loss_weight * loss_fm + \
            self.multi_mel_spectral_recon_loss_weight * loss_mel + \
            self.tpr_loss_weight * loss_tpr + loss_f0
        return {'loss': loss, 'loss_gen': loss_gen, 'loss_fm': loss_fm, 'loss_mel': loss_mel, 'loss_tpr': loss_tpr, 'loss_f0': loss_f0}

    def forward_discriminator(self, batch, device):
        real_speech = batch['speech'].to(device)
        # 1. calculate generator outputs
        with torch.no_grad():
            generated_speech, generated_f0 = self.generator(batch, device)
        # 2. calculate discriminator outputs
        y_d_rs, y_d_gs, fmap_rs, fmap_gs = self.discriminator(real_speech, generated_speech)
        # 3. calculate discriminator losses, tpr losses [Optional]
        loss_disc, _, _ = discriminator_loss(y_d_rs, y_d_gs)
        if self.tpr_loss_weight != 0:
            loss_tpr = tpr_loss(y_d_rs, y_d_gs, self.tpr_loss_tau)
        else:
            loss_tpr = torch.zeros(1).to(device)
        loss = loss_disc + self.tpr_loss_weight * loss_tpr
        return {'loss': loss, 'loss_disc': loss_disc, 'loss_tpr': loss_tpr}



import os
import random
import sys
import uuid
import json
import logging
import traceback

import grpc
from dotenv import load_dotenv
import numpy as np
import soundfile

try:
    cur_dir = os.path.dirname(__file__)
    sys.path.insert(1, os.path.join(cur_dir, "../../../common/grpc/idl"))
    from achatbot.common.grpc.idl.tts_pb2_grpc import TTSStub
    from achatbot.common.grpc.idl.tts_pb2 import (
        LoadModelRequest,
        LoadModelResponse,
        SynthesizeRequest,
        SynthesizeResponse,
        GetStreamInfoRequest,
        GetStreamInfoReponse,
        GetVoicesRequest,
        GetVoicesResponse,
        SetVoiceRequest,
        SetVoiceResponse,
    )
except ModuleNotFoundError as e:
    raise Exception(f"grpc import error: {e}")

from achatbot.common.factory import EngineClass
from achatbot.common.interface import IAudioStream
from achatbot.modules.speech.audio_stream import AudioStreamEnvInit
from achatbot.modules.speech.player import PlayerEnvInit
from achatbot.modules.speech.tts import TTSEnvInit
from achatbot.common.grpc.interceptors.authentication_client import add_authentication
from achatbot.common.logger import Logger
from achatbot.common.types import RECORDS_DIR, SessionCtx, PYAUDIO_PAFLOAT32, PYAUDIO_PAINT16
from achatbot.common.session import Session


load_dotenv(override=True)

Logger.init(
    os.getenv("LOG_LEVEL", "debug").upper(),
    app_name="chat-bot-tts-client",
    is_file=False,
    is_console=True,
)


def load_model(tts_stub: TTSStub):
    tag = os.getenv("TTS_TAG", "tts_edge")
    is_reload = bool(os.getenv("IS_RELOAD", None))
    kwargs = TTSEnvInit.map_config_func[tag]()
    request = LoadModelRequest(tts_tag=tag, is_reload=is_reload, json_kwargs=json.dumps(kwargs))
    logging.debug(request)
    response = tts_stub.LoadModel(request)
    logging.debug(response)


def synthesize_us(tts_stub: TTSStub):
    request_data = SynthesizeRequest(tts_text="hello,你好，我是机器人")
    response_iterator = tts_stub.SynthesizeUS(request_data)
    for response in response_iterator:
        yield response.tts_audio


def get_stream_info(tts_stub: TTSStub):
    response: GetStreamInfoReponse = tts_stub.GetStreamInfo(GetStreamInfoRequest())
    return response


def get_voices(tts_stub: TTSStub):
    response: GetVoicesResponse = tts_stub.GetVoices(GetVoicesRequest())
    return response.voices


def set_voice(tts_stub: TTSStub, voice: str):
    response = tts_stub.SetVoice(SetVoiceRequest(voice=voice))
    return response


logging.basicConfig(level=logging.DEBUG)

"""
TTS_TAG=tts_edge IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_g IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_coqui IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_chat IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_cosy_voice IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_fishspeech IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_f5 IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_openvoicev2 IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_kokoro IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_onnx_kokoro IS_RELOAD=1 KOKORO_ESPEAK_NG_LIB_PATH=/usr/local/lib/libespeak-ng.1.dylib KOKORO_LANGUAGE=cmn python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_cosy_voice2 \
    COSY_VOICE_MODELS_DIR=./models/FunAudioLLM/CosyVoice2-0.5B \
    COSY_VOICE_REFERENCE_AUDIO_PATH=./test/audio_files/asr_example_zh.wav \
    IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
TTS_TAG=tts_llasa IS_SAVE=1 IS_RELOAD=1 python -m achatbot.cmd.grpc.speaker.client
"""
if __name__ == "__main__":
    player = None
    channel = None
    is_save = bool(os.getenv("IS_SAVE", ""))
    tts_tag = os.getenv("TTS_TAG")
    res = bytearray()
    try:
        client_id = str(uuid.uuid4())
        session = Session(**SessionCtx(client_id).__dict__)
        # todo: up to the rpc gateway to auth
        token = "oligei-tts"
        authentication = add_authentication("authorization", token)
        port = os.getenv("PORT", "50052")
        channel = grpc.insecure_channel(f"localhost:{port}")
        channel = grpc.intercept_channel(channel, authentication)
        tts_stub = TTSStub(channel)

        load_model(tts_stub)

        stream_info = get_stream_info(tts_stub)
        logging.debug(f"stream_info:{stream_info}")

        voices = get_voices(tts_stub)
        logging.debug(f"voices:{voices}")

        # voice must match language to select
        # voice = random.choice(voices)
        # logging.debug(f"set voice:{voice}")
        # set_voice(tts_stub, voice)

        tts_audio_iter = synthesize_us(tts_stub)

        if is_save is False:
            audio_out_stream: IAudioStream | EngineClass = (
                AudioStreamEnvInit.initAudioOutStreamEngine()
            )
            player = PlayerEnvInit.initPlayerEngine()
            player.set_out_stream(audio_out_stream)
            player.open()
            player.start(session)

        for tts_audio in tts_audio_iter:
            logging.debug(f"play tts_chunk len:{len(tts_audio)}")
            session.ctx.state["tts_chunk"] = tts_audio
            if is_save is False:
                player.play_audio(session)
            else:
                res.extend(tts_audio)

        if is_save is False:
            player.stop(session)
    except grpc.RpcError as e:
        logging.error(f"grpc.RpcError: {e}")
    except Exception as e:
        tb_str = traceback.format_exc()
        logging.error(f"Exception: {e}; traceback: {tb_str}")
    finally:
        if len(res) > 0:
            file_name = f"grpc_{tts_tag}.wav"
            os.makedirs(RECORDS_DIR, exist_ok=True)
            file_path = os.path.join(RECORDS_DIR, file_name)
            np_dtype = np.float32
            if stream_info.format == PYAUDIO_PAINT16:
                np_dtype = np.int16
            data = np.frombuffer(res, dtype=np_dtype)
            soundfile.write(file_path, data, stream_info.rate)
            logging.info(f"save audio stream to {file_path}")

        if is_save is False:
            channel and channel.close()
            player and player.close()
